import threading
from .. import Progress

__all__ = ["MultiProcess"]


class MultiProcess:
    """
    MultiProcess is a utility class designed to execute a given function in parallel using multiple threads.

    Attributes:
        max_threads (int): Maximum number of threads to be used for parallel processing.
        _active_threads (int): The number of currently running threads.
        _lock (threading.Lock): A lock to manage access to shared resources.
        _thread_available (threading.Condition): A condition variable associated with `lock` to manage thread synchronization.
        _terminate (bool): A flag indicating if the thread execution should be terminated prematurely.
    """

    def __init__(self, max_threads: int):
        """
        Initializes a new instance of the MultiProcess class.

        Args:
            max_threads (int): The maximum number of threads to be used.
        """
        self.max_threads = max_threads
        self._active_threads = 0
        self._lock = threading.Lock()
        self._thread_available = threading.Condition(self._lock)
        self._terminate = False
        self._results = []

    def execute(self, function, values, *args, **kwargs) -> list:
        """
        Execute the given function using multiple threads on the provided values.

        Args:
            function (Callable): The function to be executed.
            values (Iterable): A list or other iterable of values on which the function will be executed.

        @return: ordered each given function returns
        """
        for indx, value in enumerate(
                Progress.prog(values, custom_state_func=lambda: f'Threads Running: {self._active_threads}')):
            self.wait_for_available_thread()
            if self._terminate:
                print("Terminating due to an exception in one of the threads. Returning processed data...")
                break
            thread = threading.Thread(target=self._execute_function_thread, name=f'Thread-{indx}',
                                      args=(function, value, indx, args, kwargs))
            with self._lock:
                self._active_threads += 1
            thread.start()
        self._terminate = False
        return self._get_results()

    def _get_results(self):
        with self._thread_available:
            while self._active_threads > 0:
                self._thread_available.wait()
            results = [val[-1] for val in sorted(self._results, key=lambda val: val[0])]
            self._results = []
            return results

    def wait_for_available_thread(self):
        """
        Blocks the current thread until a thread becomes available or the maximum thread limit is not reached.
        """
        with self._thread_available:
            while self._active_threads >= self.max_threads:
                self._thread_available.wait()

    def _execute_function_thread(self, function, value, indx, args, kwargs):
        """
        Internal method to execute the function on the given value. Handles exceptions and manages the active thread count.

        Args:
            function (Callable): The function to be executed.
            value: A single value from the values iterable.
        """
        try:
            if not self._terminate:
                self._results.append((indx, function(value, *args, **kwargs)))
        except Exception as e:
            print(f"Error encountered in thread: {e}")
            self._terminate = True
        finally:
            with self._thread_available:
                self._active_threads -= 1
                self._thread_available.notify()



"""
Copyright (c) 2019 The Cereja Project
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import ast
import ctypes
import gc
import math
import threading
import time
from collections import OrderedDict, defaultdict
from importlib import import_module
import importlib
import sys
import types
import random
from typing import Any, Union, List, Tuple, Sequence, Iterable, Dict, MappingView, Optional, Callable
import logging
import itertools
from copy import copy
import inspect
from pprint import pprint

from .decorators import time_exec, depreciation
# Needed init configs
from ..config.cj_types import ClassType, FunctionType, Number
from itertools import combinations as itertools_combinations

__all__ = [
    "CjTest",
    "camel_to_snake",
    "camel_case_to_snake",
    "snake_case_to_camel",
    "combine_with_all",
    "fill",
    "get_attr_if_exists",
    "get_implements",
    "get_instances_of",
    "import_string",
    "install_if_not",
    "invert_dict",
    "logger_level",
    "module_references",
    "set_log_level",
    "time_format",
    "string_to_literal",
    "rescale_values",
    "Source",
    "sample",
    "visualize_sample",
    "obj_repr",
    "truncate",
    "type_table_of",
    "list_methods",
    "can_do",
    "chunk",
    "is_iterable",
    "is_indexable",
    "is_sequence",
    "is_numeric_sequence",
    "clipboard",
    "sort_dict",
    "dict_append",
    "to_tuple",
    "dict_to_tuple",
    "list_to_tuple",
    "group_by",
    "dict_values_len",
    "dict_max_value",
    "dict_min_value",
    "dict_filter_value",
    "get_zero_mask",
    "get_batch_strides",
    "Thread",
    "prune_values",
    "split_sequence",
    "has_length",
    "combinations",
    "combinations_sizes",
    "value_from_memory"
]

logger = logging.getLogger(__name__)


# TODO: Remove Thread class, check cereja.concurrently.process.MultiProcess implementation
class Thread(threading.Thread):
    def __init__(
            self, target, args=None, kwargs=None, name=None, daemon=None, callback=None
    ):
        # while threading.active_count() > os.cpu_count() * 2:
        #    time.sleep(0.1)
        super().__init__(daemon=daemon, name=name)
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        self._func = target
        self._args = args
        self._kwargs = kwargs
        self._callback = callback

    def run(self):
        res = self._func(*self._args, **self._kwargs)
        if self._callback:
            self._callback(res)


def is_indexable(v):
    return hasattr(v, "__getitem__")


def split_sequence(seq, is_break_fn):
    """
    Divide a sequence into sub-sequences at the point defined by a given is_break_fn function.

    Args:
        seq (list): The sequence to be divided into sub-sequences.
        is_break_fn (callable): A function that takes two elements from the sequence and
            returns True if there is a break between them and False otherwise.

    Returns:
        list: A list of sub-sequences, each containing a sequence of consecutive elements.

    Examples:
        >>> seq = [1, 2, 3, 4, 5, 2, 3, 4, 5, 6, 7, 8]
        >>> is_even = lambda x, y: x % 2 == 0 and y % 2 == 0
        >>> sub_seqs = split_sequence(seq, is_even)
        >>> sub_seqs
        [[1, 2, 3, 4, 5], [2, 3, 4, 5, 6, 7, 8]]
    """
    sub_seqs = []
    sub_seq = [seq[0]]
    for i in range(1, len(seq)):
        if is_break_fn(seq[i - 1], seq[i]):
            sub_seqs.append(sub_seq)
            sub_seq = []
        sub_seq.append(seq[i])
    sub_seqs.append(sub_seq)
    return sub_seqs


def chunk(
        data: Sequence,
        batch_size: int = None,
        fill_with: Any = None,
        is_random: bool = False,
        max_batches: int = None,
) -> List[Union[Sequence, List, Tuple, Dict]]:
    """

    e.g:
    >>> import cereja as cj

    >>> data = list(range(15))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]

    >>> cj.chunk(data, batch_size=4)
    [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14]]

    >>> cj.chunk(data, batch_size=4, is_random=True, fill_with=0)
    [[7, 2, 11, 4], [10, 6, 1, 13], [12, 9, 5, 0], [8, 3, 14, 0]]

    >>> data = {"key1": 'value1', "key2": 'value2', "key3": 'value3', "key4": 'value4'}

    >>> cj.chunk(data, batch_size=2,is_random=True)
    [{'key3': 'value3', 'key2': 'value2'}, {'key1': 'value1', 'key4': 'value4'}]

    @param data: Iterable data
    @param batch_size: number of items per batch
    @param fill_with: Any, but isn't valid for dict
    @param is_random: shuffle data
    @param max_batches: limit number of batches
    @return: list of batches
    """

    assert (
            is_iterable(data) and len(data) > 0
    ), f"Chunk isn't possible, because value {data} isn't valid."
    if batch_size is None and max_batches is None:
        return [data]

    # used to return the same data type
    __parser = None

    if isinstance(data, (dict, tuple, set)):
        __parser = type(data)
        data = data.items() if isinstance(data, dict) else data

    data = (
        list(data)
        if isinstance(data, (set, tuple, str, bytes, bytearray, MappingView))
        else copy(data)
    )
    if not batch_size or batch_size > len(data) or batch_size < 1:
        if isinstance(max_batches, (int, float)) and max_batches > 0:
            batch_size = math.ceil(len(data) / max_batches)
        else:
            batch_size = len(data)

    if is_random:
        random.shuffle(data)

    if max_batches is None:
        max_batches = (
            len(data) // batch_size
            if len(data) % batch_size == 0
            else len(data) // batch_size + 1
        )

    batches = []
    for i in range(0, len(data), batch_size):

        result = data[i: i + batch_size]
        if fill_with is not None and len(result) < batch_size:
            result += [fill_with] * (batch_size - len(result))
        batches.append(__parser(result) if __parser is not None else result)
        max_batches -= 1
        if not max_batches:
            break
    return batches


def _get_tkinter():
    try:
        from tkinter import Tk
    except ImportError:
        raise ValueError("Sorry. Isn't possible.")
    return Tk()


def clipboard() -> str:
    return _get_tkinter().clipboard_get()


def truncate(data: Union[Sequence], k_iter: int = 0, k_str: int = 0, k_dict_keys: int = 0):
    """
    Truncates the data to the specified number of elements in half(+1 if odd) + half, adding a filler in between.
    If the data is a dictionary, then truncates recursively until data is "truncatable".
    Allows to specify a different to k to strings and number of dictionary keys.
    Args:
        data: sequence
        k_iter: number of elements of iterable data to return
        k_str: number of elements of strings to return
        k_dict_keys: number of dictionary keys to return
    Returns:
        truncated data
    """
    assert all(isinstance(k, int) and k >= 0 for k in
               (k_iter, k_str, k_dict_keys)), 'k parameters should be an integer equal to or larger than 0'

    if isinstance(data, dict):
        if k_dict_keys:
            data = {key: data.get(key, '<…>') for key in truncate(list(data.keys()), k_dict_keys)}
            if k_iter or k_str:
                data = {truncate(key, k_iter, k_str): value for key, value in data.items()}
        for key, value in data.items():
            data[key] = truncate(value, k_iter, k_str, k_dict_keys)
    elif isinstance(data, str):
        if k_str == 0:
            return data
        k_str = min(k_str, len(data))
        n = k_str // 2
        filler = '…' if len(data) > k_str else ''
        return data[:(k_str - n)] + filler + data[-n:] if n else data[:(k_str - n)] + filler
    elif isinstance(data, Iterable):
        if k_iter == 0:
            return data
        k_iter = min(k_iter, len(data))
        n = k_iter // 2
        if isinstance(data, bytes):
            filler = b'...' if len(data) > k_iter else b''
            return data[:(k_iter - n)] + filler + data[-n:] if n else data[:(k_iter - n)] + filler
        else:
            filler = ['<…>'] if len(data) > k_iter else []
        if isinstance(data, list):
            if n:
                return [truncate(dt, k_iter, k_str, k_dict_keys) for dt in data[:(k_iter - n)]] + filler + \
                    [truncate(dt, k_iter, k_str, k_dict_keys) for dt in data[-n:]]
            return [truncate(dt, k_iter, k_str, k_dict_keys) for dt in data[:(k_iter - n)]] + filler
        elif isinstance(data, set):
            if n:
                return set([truncate(dt, k_iter, k_str, k_dict_keys) for dt in list(data)[:(k_iter - n)]] + filler + \
                           [truncate(dt, k_iter, k_str, k_dict_keys) for dt in list(data)[-n:]])
            return set([truncate(dt, k_iter, k_str, k_dict_keys) for dt in list(data)[:(k_iter - n)]] + filler)
        return data
    return data


def obj_repr(
        obj_, attr_limit=10, val_limit=3, show_methods=False, show_private=False, deep=3
):
    try:
        if isinstance(obj_, (str, bytes)):
            return truncate(obj_, k_iter=attr_limit)
        if isinstance(obj_, (bool, float, int, complex)):
            return obj_
        rep_ = []
        if deep > 0:
            for attr_ in dir(obj_):
                if attr_.startswith("_") and not show_private:
                    continue
                obj = obj_.__getattribute__(attr_)

                if isinstance(obj, (str, bool, float, int, complex, bytes, bytearray)):
                    rep_.append(f"{attr_} = {obj_repr(obj)}")
                    continue
                if callable(obj) and not show_methods:
                    continue

                if is_iterable(obj):
                    temp_v = []
                    for k in obj:
                        if isinstance(obj, dict):
                            k = f"{k}:{type(obj[k])}"
                        elif is_iterable(k):
                            k = obj_repr(k, deep=deep)
                            deep -= 1
                        else:
                            k = str(k)
                        temp_v.append(k)
                        if len(temp_v) == val_limit:
                            break
                    temp_v = ", ".join(temp_v)  # fix me, if bytes ...
                    obj = f"{obj.__class__.__name__}({temp_v} ...)"
                rep_.append(f"{attr_} = {obj}")
                if len(rep_) >= attr_limit:
                    rep_.append("...")
                    break
        else:
            return repr(obj_)

    except Exception as err:
        logger.error(err)
        rep_ = []
    rep_ = ",\n    ".join(rep_)
    __repr_template = f"""
    {rep_}
    """
    return f"{obj_.__class__.__name__} ({__repr_template})"


def can_do(obj: Any) -> List[str]:
    """
    List methods and attributes of a Python object.

    It is essentially the builtin `dir` function without the private methods and attributes

    @param obj: Any
    @return: list of attr names sorted by name
    """
    return sorted([i for i in filter(lambda attr: not attr.startswith("_"), dir(obj))])


def sample(
        v: Sequence, k: int = None, is_random: bool = False) -> Union[list, dict, set, Any]:
    """
    Get sample of anything

    @param v: Any
    @param k: int
    @param is_random: default False
    @return: sample iterable
    """
    result = chunk(v, batch_size=k, is_random=is_random, max_batches=1)[0]

    if k == 1 and len(result) == 1:
        if isinstance(result, dict):
            return result
        return next(iter(result))
    return result


def visualize_sample(v: Sequence, k: int = None, is_random: bool = False, tr_k_iter: int = 6, tr_k_str: int = 20,
                     tr_k_dict_keys: int = 20, p_print: bool = True):
    """
    Samples then (p)prints a (truncated) version of the sample. Helpful for visualizing data structures.
    Args:
        v: sequence
        k: number of samples
        is_random: should shuffle
        tr_k_iter: how many items of iterables to keep in the truncated version of the sample, `0` to disable
        tr_k_str: truncated length of strings, `0` to disable
        tr_k_dict_keys: truncated number of dictionary keys, `0` to disable
        p_print: should pprint or print
    """

    obj_sample = sample(v=copy(v), k=k, is_random=is_random)
    truncated_sample = truncate(data=obj_sample, k_iter=tr_k_iter, k_str=tr_k_str, k_dict_keys=tr_k_dict_keys)
    if p_print:
        pprint(truncated_sample)
    else:
        print(truncated_sample)


def type_table_of(o: Union[list, tuple, dict]):
    if isinstance(o, (list, tuple)):
        type_table = {i: type(i) for i in o}
    elif isinstance(o, dict):
        type_table = {}
        for k, v in o.items():
            if isinstance(o, dict):
                v = type_table_of(v)
            type_table[k] = (v, type(v))
    else:
        type_table = {o: type(o)}

    return type_table


def camel_case_to_snake(value: str):
    snaked_ = []
    for i, char in enumerate(value):
        if not i == 0 and char.isupper():
            char = f"_{char}"
        snaked_.append(char)
    return "".join(snaked_).lower()


@depreciation(alternative="camel_case_to_snake")
def camel_to_snake(value: str):
    return camel_case_to_snake(value)


def camel_case(value: str, sep: Union[str, None] = None, upper_first=False):
    if sep is None:
        sep = [" "]
    elif isinstance(sep, (str, int, float)):
        sep = [str(sep)]
    value = value.lower().title()
    for sp in sep:
        value = value.replace(sp, '')
    if upper_first:
        return value
    return value[0].lower() + value[1:]


def snake_case_to_camel(value, upper_first=False):
    return camel_case(value, sep="_", upper_first=upper_first)


def get_implements(klass: type):
    classes = klass.__subclasses__()
    collected_classes = []
    for k in classes:
        k_classes = k.__subclasses__()
        if k_classes:
            collected_classes += get_implements(k)
        if not k.__name__.startswith("_"):
            collected_classes.append(k)
    return collected_classes


def get_instances_of(klass: type):
    return filter(lambda x: isinstance(x, klass), gc.get_objects())


def _invert_parser_key(key):
    return to_tuple(key) if isinstance(key, (list, set, dict)) else key


def _invert_append(obj, k, v):
    dict_append(obj, k, v)
    if len(obj[k]) == 1:
        obj[k] = obj[k][0]


def invert_dict(dict_: Union[dict, set]) -> dict:
    """
    Inverts the key by value
    e.g:
    >>> example = {"a": "b", "c": "d"}
    >>> invert_dict(example)
    {"b" : "a", "d": "c"}
    :return: dict
    """

    if not isinstance(dict_, dict):
        raise TypeError("Send a dict object.")
    new_dict = {}
    for key, value in dict_.items():
        key = _invert_parser_key(key)

        if isinstance(value, dict):
            if key not in new_dict:
                new_dict.update({key: invert_dict(value)})
            else:
                _invert_append(new_dict, key, invert_dict(value))
            continue
        if isinstance(value, (tuple, list, set)):
            for k in dict_[key]:
                k = _invert_parser_key(k)
                _invert_append(new_dict, k, key)
            continue

        if value not in new_dict:
            new_dict[value] = key
        else:
            value = _invert_parser_key(value)
            _invert_append(new_dict, value, key)

    return new_dict


def group_by(values, fn, get_freq: bool = False) -> Union[dict, Tuple[dict, dict]]:
    """
    group items by result of fn (function)

    eg.
    >>> import cereja as cj
    >>> values = ['joab', 'leite', 'da', 'silva', 'Neto', 'você']
    >>> cj.group_by(values, lambda x: 'N' if x.lower().startswith('n') else 'OTHER')
    # {'OTHER': ['joab', 'leite', 'da', 'silva', 'você'], 'N': ['Neto']}
    >>> cj.group_by(values, lambda x: 'N' if x.lower().startswith('n') else 'OTHER', get_freq=True)
    # ({'OTHER': ['joab', 'leite', 'da', 'silva', 'você'], 'N': ['Neto']},
    #  {'OTHER': 5, 'N': 1})

    @param values: list of values
    @param fn: a function
    @param get_freq: if True, returns a tuple containing the main result of the function followed by the frequency
    distribution of the groups, respectively
    """
    d = defaultdict(list)
    for el in values:
        d[fn(el)].append(el)

    d = dict(d)
    result = (d, dict(map(lambda t: (t[0], len(t[1])), d.items()))) if get_freq else d
    return result


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
                f"Module {module_path} does not define a {class_name} attribute/class"
        ) from err


def get_attr_if_exists(obj: Any, attr: str) -> Union[object, None]:
    if hasattr(obj, attr):
        return getattr(obj, attr)
    return None


def time_format(seconds: float, format_="%H:%M:%S") -> Union[str, float]:
    """
    Default format is '%H:%M:%S'

    >>> time_format(3600)
    '01:00:00'

    """
    # this because NaN
    if seconds >= 0 or seconds < 0:
        time_ = time.strftime(format_, time.gmtime(abs(seconds)))
        if seconds < 0:
            return f"-{time_}"
        return time_
    return seconds  # NaN


def fill(value: Union[list, str, tuple], max_size, with_=" ") -> Any:
    """
    Calculates and adds value
    """
    fill_values = [with_] * (max_size - len(value))
    if isinstance(value, str):
        fill_values = " ".join(fill_values)
        value = f"{value}{fill_values}"
    elif isinstance(value, list):
        value += fill_values
    elif isinstance(value, tuple):
        value += tuple(fill_values)
    return value


def list_methods(klass) -> List[str]:
    methods = []
    for i in dir(klass):
        if i.startswith("_") or not callable(getattr(klass, i)):
            continue
        methods.append(i)
    return methods


def string_to_literal(val: Union[str, bytes]):
    if isinstance(val, (str, bytes)):
        try:
            return ast.literal_eval(val)
        except:
            pass
    return val


def module_references(instance: types.ModuleType, **kwargs) -> dict:
    """
    dict of all functions and classes defined in the module.
    To also list the variables it is necessary to define explicitly with the special variable on your module
    _include
    **kwargs:
    _include -> to includes any definition and variables
    _exclude -> to exclude any definition
    :param instance:
    :return: List[str]
    """
    assert isinstance(
            instance, types.ModuleType
    ), "You need to submit a module instance."
    logger.debug(f"Checking module {instance.__name__}")
    definitions = {}
    for i in dir(instance):
        if i.startswith("_"):
            continue
        exclude = (
                get_attr_if_exists(instance, "_exclude") or kwargs.get("_exclude") or []
        )
        include = (
                get_attr_if_exists(instance, "_include") or kwargs.get("_include") or []
        )

        obj = get_attr_if_exists(instance, i)

        if i in include:
            definitions[i] = obj

        if obj is not None and i not in exclude and callable(obj):
            if obj.__module__ == instance.__name__:
                definitions[i] = obj
    logger.debug(f"Collected: {definitions}")
    return definitions


def get_python_executable():
    import cereja as cj
    if sys.platform == 'win32':
        exec_paths = ['python.exe',
                      'Scripts/python.exe',
                      'bin/python.exe']
        exec_python = None
        for p in exec_paths:
            _exec_python = cj.Path(f'{sys.exec_prefix}/{p}')
            if _exec_python.exists:
                exec_python = _exec_python
                break
        assert exec_python is not None, 'Error to install dependences: Python executable not founded.'

    else:
        exec_python = sys.executable

    return exec_python


def install_if_not(lib_name: str):
    from cereja.display import console

    try:
        importlib.import_module(lib_name)
        output = "Alredy Installed"
    except ImportError:
        from ..system.commons import run_on_terminal

        command_ = f'"{get_python_executable()}" -m pip install {lib_name}'
        output = run_on_terminal(command_)
    console.log(output)


def set_log_level(level: Union[int, str]):
    """
    Default log level is INFO
    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10
    NOTSET = 0
    """
    log = logging.getLogger()
    log.setLevel(level)
    logger.info(f"Update log level to {level}")


def logger_level():
    import logging

    return logging.getLogger().level


def combine_with_all(
        a: list, b: list, n_a_combinations: int = 1, is_random: bool = False
) -> List[Tuple[Any, ...]]:
    """
    >>> a = [1, 2, 3]
    >>> b = ['anything_a', 'anything_b']
    >>> combine_with_all(a, b)
    [(1, 'anything_a'), (1, 'anything_b'), (2, 'anything_a'), (2, 'anything_b'), (3, 'anything_a'), (3, 'anything_b')]
    >>> combine_with_all(a, b, n_a_combinations=2)
    [((1, 2), 'anything_a'), ((1, 2), 'anything_b'),
    ((1, 3), 'anything_a'), ((1, 3), 'anything_b'),
    ((2, 3), 'anything_a'), ((2, 3), 'anything_b')]
    """
    if not isinstance(n_a_combinations, int):
        raise TypeError(f"Please send {int}.")
    n_a_combinations = len(a) if n_a_combinations > len(a) else abs(n_a_combinations)

    combination = (
        itertools.combinations(a, n_a_combinations) if n_a_combinations > 1 else a
    )
    product_with_b = list(itertools.product(combination, b))
    if is_random:
        random.shuffle(product_with_b)
    return product_with_b


def value_from_memory(memory_id):
    """
    Retrieve the Python object stored at a specific memory address.

    Args:
    memory_id (int): The memory address of the object.

    Returns:
    object: The Python object stored at the given memory address, or None if the retrieval fails.
    """
    try:
        return ctypes.cast(memory_id, ctypes.py_object).value
    except (ValueError, ctypes.ArgumentError):
        raise ValueError(f"Memory ID {memory_id} isn't valid.")


def sort(iterable, reverse=False):
    """
    Sort a list that may contain a mix of different types including other lists or tuples.

    Args:
    iterable (list): A list which may contain mixed types.

    Returns:
    list: A sorted list.
    """

    def sort_key(elem):
        """
        A key function for sorting which handles elements of different types.
        Converts the element to a string for comparison purposes.
        """
        if isinstance(elem, (list, tuple)):
            return str([sort_key(e) for e in elem])
        return str(elem)

    return sorted(iterable, key=sort_key, reverse=reverse)


def combinations(iterable, size, is_sorted=False):
    """
    Generate all possible combinations of a certain size from an iterable.

    Args:
    iterable (iterable): An iterable of Python objects.
    size (int): The size of each combination.

    Returns:
    list of tuples: A list containing tuples of the combinations generated.
    """
    if not is_sorted:
        return list(itertools_combinations(iterable, size))
    try:
        return list(map(lambda x: tuple(sort(x)), itertools_combinations(iterable, size)))
    except Exception as err:
        raise Exception(f"Can't sort the pairs. {err}")


def combinations_sizes(iterable, min_size, max_size, is_sorted=False):
    """
    Generate all possible combinations for all sizes within the specified range from an iterable.

    Args:
    iterable (iterable): An iterable of Python objects.
    min_size (int): The minimum size of the combinations.
    max_size (int): The maximum size of the combinations.

    Returns:
    list of tuples: A list containing tuples of all combinations for sizes between min_size and max_size.
    """
    res = []
    for n in range(min_size, max_size + 1):
        res.extend(combinations(iterable, n, is_sorted=is_sorted))
    return res


class CjTest(object):
    __template_unittest_function = """
    def test_{func_name}(self):
        pass
        """
    __template_unittest_class = """
class {class_name}Test(unittest.TestCase):
    {func_tests}
        """

    __template_unittest = """import unittest

{tests}

if __name__ == '__main__':
    unittest.main()

        """

    __prefix_attr_err = "Attr Check Error {attr_}."

    def __init__(self, instance_obj: object):
        self._prefix_attr = f"__{instance_obj.__class__.__name__}__"
        self._instance_obj = instance_obj
        self._set_attr_current_values()
        self._checks = []
        self._n_checks_passed = 0

    @property
    def checks(self):
        return self._checks

    @property
    def n_checks(self):
        return len(self._checks)

    @property
    def _instance_obj_attrs(self):
        return filter(
                lambda attr_: attr_.__contains__("__") is False, dir(self._instance_obj)
        )

    def _get_attr_obj(self, attr_: str):
        if not hasattr(self._instance_obj, attr_):
            raise ValueError(f"Attr {attr_} not found.")
        value = getattr(self._instance_obj, attr_)
        return self._Attr(attr_, value)

    def _set_attr_current_values(self):
        for attr_ in self._instance_obj_attrs:
            attr_obj = self._get_attr_obj(attr_)
            attr_name = self.parse_attr(attr_)
            setattr(self, attr_name, attr_obj)

    def parse_attr(self, attr_: str):
        attr_ = self._valid_attr(attr_)
        return f"{self._prefix_attr}{attr_}"

    def __getattr__(self, item):
        return self.__getattribute__(self.parse_attr(item))

    class _Attr(object):
        def __init__(self, name: str, value: Any):
            self.name = name
            self.is_callable = callable(value)
            self.is_private = self.name.startswith("_")
            self.is_bool = value is True or value is False
            self.is_class = isinstance(value, ClassType)
            self.is_function = isinstance(value, FunctionType)
            self.class_of_attr = value.__class__
            self._operator_repr = None
            self.tests_case = []

        def __repr__(self):
            return f"{self.name}"

        def __str__(self):
            return f"{self.name}"

        def __len__(self):
            return len(self.tests_case)

        def __eq__(self, other):
            """==value"""
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__eq__, "==")
            return self

        def __ge__(self, other):
            """>=value"""
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__ge__, ">=")
            return self

        def __gt__(self, other):
            """>value"""
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__gt__, ">")
            return self

        def __le__(self, other):
            """<=value."""
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__le__, "<=")
            return self

        def __lt__(self, other):
            """<value."""
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__lt__, "<")
            return self

        def __ne__(self, other):
            """!=value."""
            if isinstance(other, self.__class__):
                return NotImplemented
            self.tests_case = (other, self.class_of_attr.__ne__, "!=")
            return self

        def copy(self):
            return copy(self)

        def run(self, current_value):
            expected, operator, _ = self.tests_case
            if not operator(current_value, expected):
                return [f"{repr(current_value)} not {_} {repr(expected)}"]

            return []

    def _valid_attr(self, attr_name: str):
        assert hasattr(
                self._instance_obj, attr_name
        ), f"{self.__prefix_attr_err.format(attr_=repr(attr_name))} isn't defined."
        return attr_name

    def add_check(self, *check_: _Attr):
        for i in check_:
            if i not in self._checks:
                self._checks.append(i)

    def remove_check(self, index: int):
        self._checks.pop(index)

    def check_attr(self, attr_name: Union[str, _Attr]):
        if isinstance(attr_name, str):
            stored_test = self.__getattribute__(self.parse_attr(attr_name))
        else:
            stored_test = attr_name
        current_value = getattr(self._instance_obj, stored_test.name)
        if not stored_test.is_callable:
            tests_ = stored_test.run(current_value)
            passed = not any(tests_)
            self._n_checks_passed += len(stored_test) - len(tests_)
            msg_err = f"{self.__prefix_attr_err.format(attr_=repr(stored_test.name))} {' '.join(tests_)}"
            assert passed, msg_err

    def check_all(self):
        for attr_ in self._checks:
            self.check_attr(attr_)

    @classmethod
    def _get_class_test(cls, ref):
        func_tests = "".join(
                cls.__template_unittest_function.format(func_name=i)
                for i in list_methods(ref)
        )
        return cls.__template_unittest_class.format(
                class_name=ref.__name__, func_tests=func_tests
        )

    @classmethod
    def _get_func_test(cls, ref):
        return cls.__template_unittest_function.format(func_name=ref.__name__)

    @classmethod
    def _get_test(cls, ref):
        if isinstance(ref, (FunctionType, types.MethodType)):
            return cls._get_func_test(ref)
        if isinstance(ref, type):
            return cls._get_class_test(ref)
        raise TypeError("send a function or class reference")

    @classmethod
    def build_test(cls, reference):
        module_func_test = []
        tests = []
        if isinstance(reference, types.ModuleType):
            for _, ref in module_references(reference).items():
                if isinstance(ref, type):
                    tests.append(cls._get_test(ref))
                    continue
                module_func_test.append(cls._get_test(ref))
        else:
            if isinstance(reference, type):
                tests.append(cls._get_test(reference))
            else:
                module_func_test.append(cls._get_test(reference))
        if module_func_test:
            module_func_test = "".join(module_func_test)
            tests = [
                        cls.__template_unittest_class.format(
                                class_name="Module", func_tests=module_func_test
                        )
                    ] + tests
        return cls.__template_unittest.format(tests="\n".join(tests))


def _add_license(base_dir, ext=".py"):
    from cereja.file import FileIO
    from cereja.config import BASE_DIR

    licence_file = FileIO.load(BASE_DIR)
    for file in FileIO.load_files(base_dir, ext=ext, recursive=True):
        if "Copyright (c) 2019 The Cereja Project" in file.string:
            continue
        file.insert('"""\n' + licence_file.string + '\n"""')
        file.save(exist_ok=True)


def _rescale_down(input_list, size):
    assert len(input_list) >= size, f"{len(input_list), size}"

    skip = len(input_list) // size
    for n, i in enumerate(range(0, len(input_list), skip), start=1):
        if n > size:
            break
        yield input_list[i]


def _rescale_up(values, k, fill_with=None, filling="inner"):
    size = len(values)
    assert size <= k, f"Error while resizing: {size} < {k}"

    clones = math.ceil(abs(size - k) / size)
    refill_values = abs(k - size * clones)
    if filling == "pre":
        for i in range(abs(k - size)):
            yield fill_with if fill_with is not None else values[0]

    for value in values:
        # guarantees that the original value will be returned
        yield value

        if filling != "inner":
            continue

        for i in range(clones - 1):  # -1 because last line.
            # value original or fill_with.
            yield fill_with if fill_with is not None else value
        if refill_values > 0:
            refill_values -= 1
            yield fill_with if fill_with is not None else value
        k -= 1
        if k < 0:
            break
    if filling == "post":
        for i in range(abs(k - size)):
            yield fill_with if fill_with is not None else values[-1]


def _interpolate(values, k):
    if isinstance(values, list):
        from ..array import Matrix

        # because true_div ...
        values = Matrix(values)
    size = len(values)

    first_position = 0
    last_position = size - 1
    step = (last_position - first_position) / (k - 1)

    positions = [first_position]
    previous_position = positions[-1]
    for _ in range(k - 2):
        positions.append(previous_position + step)
        previous_position = positions[-1]
    positions.append(last_position)

    for position in positions:
        previous_position = math.floor(position)
        next_position = math.ceil(position)
        if previous_position == next_position:
            yield values[previous_position]
        else:
            delta = position - previous_position
            yield values[previous_position] + (
                    values[next_position] - values[previous_position]
            ) / (next_position - previous_position) * delta


def rescale_values(
        values: List[Any],
        granularity: int,
        interpolation: bool = False,
        fill_with=None,
        filling="inner",
) -> List[Any]:
    """
    Resizes a list of values
    eg.
        >>> import cereja as cj
        >>> cj.rescale_values(values=list(range(100)),granularity=12)
        [0, 8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88]
        >>> cj.rescale_values(values=list(range(5)),granularity=10)
        [0, 0, 1, 1, 2, 2, 3, 3, 4, 4]
        >>> cj.rescale_values(values=list(range(5)),granularity=10, filling='pre')
        [0, 0, 0, 0, 0, 0, 1, 2, 3, 4]
        >>> cj.rescale_values(values=list(range(5)),granularity=10, filling='post')
        [0, 1, 2, 3, 4, 4, 4, 4, 4, 4]

        @note if you don't send any value for filling a value will be chosen arbitrarily depending on the filling type.

    If interpolation is set to True, then the resized values are calculated by interpolation,
    otherwise they are sub- ou upsampled from the original list


    @param values: Sequence of anything
    @param granularity: is a integer
    @param interpolation: is a boolean
    @param fill_with: only scale up, send any value for filling
    @param filling: in case of scale up, you can define how the filling will be (pre, inner, post). 'inner' is default.
    @return: rescaled list of values.
    """

    if interpolation:
        result = list(_interpolate(values, granularity))
    else:
        if len(values) >= granularity:
            result = list(_rescale_down(values, granularity))
        else:
            result = list(
                    _rescale_up(values, granularity, fill_with=fill_with, filling=filling)
            )

    assert (
            len(result) == granularity
    ), f"Error while resizing the list size {len(result)} != {granularity}"
    return result


class Source:
    def __init__(self, reference: Any):
        self._name = None
        self._doc = inspect.getdoc(reference)
        self._source_code = inspect.getsource(reference)
        if hasattr(reference, "__name__"):
            self._name = reference.__name__

    @property
    def source_code(self):
        return self._source_code.lstrip()

    @property
    def name(self):
        return self._name

    @property
    def doc(self):
        return self._doc

    def save(self, path_, **kwargs):
        from cereja import FileIO, Path

        path_ = Path(path_)
        if path_.is_dir:
            path_ = path_.join(f"{self.name}.py")
        assert path_.suffix == ".py", "Only python source code."
        FileIO.create(path_, self._source_code).save(**kwargs)


def is_iterable(obj: Any) -> bool:
    """
    Return whether an object is iterable or not.

    This function checks if the object has an __iter__ method or supports
    sequence-like indexing via __getitem__.

    Parameters:
    obj (Any): Any object to check for iterability.

    Returns:
    bool: True if the object is iterable, False otherwise.
    """
    return hasattr(obj, '__iter__') or hasattr(obj, '__getitem__')


def has_length(seq):
    try:
        _ = len(seq)
        return True
    except TypeError:
        return False


def is_sequence(obj: Any) -> bool:
    """
    Return whether an object a Sequence or not, exclude strings and empty obj.

    :param obj: Any object for check
    """
    if isinstance(obj, (str, dict, bytes, int, float)):
        return False
    return is_iterable(obj)


def is_numeric_sequence(obj: Sequence[Number]) -> bool:
    try:
        from cereja.array import flatten

        sum(flatten(obj))
    except (TypeError, ValueError):
        return False
    return True


def sort_dict(
        obj: dict,
        by_keys=False,
        by_values=False,
        reverse=False,
        by_len_values=False,
        func_values=None,
        func_keys=None,
) -> OrderedDict:
    func_values = (
        (lambda v: len(v) if by_len_values else v)
        if func_values is None
        else func_values
    )
    func_keys = (lambda k: k) if func_keys is None else func_keys

    key_func = None
    if (by_keys and by_values) or (not by_keys and not by_values):
        key_func = lambda x: (func_keys(x[0]), func_values(x[1]))
    elif by_keys:
        key_func = lambda x: func_keys(x[0])
    elif by_values:
        key_func = lambda x: func_values(x[1])
    return OrderedDict(sorted(obj.items(), key=key_func, reverse=reverse))


def list_to_tuple(obj):
    assert isinstance(
            obj, (list, set, tuple)
    ), f"Isn't possible convert {type(obj)} into {tuple}"
    result = []
    for i in obj:
        if isinstance(i, list):
            i = list_to_tuple(i)
        elif isinstance(i, (set, dict)):
            i = dict_to_tuple(i)
        result.append(i)
    return tuple(result)


def dict_values_len(obj, max_len=None, min_len=None, take_len=False):
    return {
        i: len(obj[i]) if take_len else obj[i]
        for i in obj
        if (max_len is None or len(obj[i]) <= max_len)
           and (min_len is None or len(obj[i]) >= min_len)
    }


def dict_to_tuple(obj):
    assert isinstance(
            obj, (dict, set)
    ), f"Isn't possible convert {type(obj)} into {tuple}"
    result = []
    if isinstance(obj, set):
        return tuple(obj)
    for k, v in obj.items():
        if isinstance(v, (dict, set)):
            v = dict_to_tuple(v)
        elif isinstance(v, list):
            v = list_to_tuple(v)
        result.append((k, v))
    return tuple(result)


def to_tuple(obj):
    if isinstance(obj, (set, dict)):
        return dict_to_tuple(obj)
    if isinstance(obj, (list, tuple)):
        return list_to_tuple(obj)
    return tuple(obj)


def dict_append(obj: Dict[Any, Union[List, Tuple]],
                key: Any,
                *v,
                unique_values: bool = False,
                sort_values_by: Optional[Callable] = None,
                reverse: bool = False) -> Dict:
    """
    Add items to a key in the dictionary. If the key doesn't exist, it's created with a list and the given values.

    e.g:

    >>> my_dict = {}
    >>> dict_append(my_dict, 'key_eg', 1,2,3,4,5,6)
    {'key_eg': [1, 2, 3, 4, 5, 6]}
    >>> dict_append(my_dict, 'key_eg', [1,2])
    {'key_eg': [1, 2, 3, 4, 5, 6, [1, 2]]}

    Parameters:
    - obj: A dictionary with list or tuple values.
    - key: The key to which the values should be added.
    - v: The values to be added.
    - unique_values: Whether to ensure the values are unique.
    - sort_values_by: Optional function to sort the values.
    - reverse: Used if sort_values_by is passed.

    Returns:
    Updated dictionary.
    """
    """
        if not isinstance(obj[key], (list, tuple)):
        obj[key] = [obj[key]]
    if isinstance(obj[key], tuple):
        obj[key] = (
            *obj[key],
            *v,
        )
    else:
        for i in v:
            obj[key].append(i)
    """

    # Ensure we're working with a dictionary
    if not isinstance(obj, dict):
        raise TypeError("Error on append values. Please provide a dictionary object.")

    # Append values to existing list or tuple, or create a new list if key doesn't exist
    if key not in obj:
        obj[key] = []

    if not isinstance(obj[key], (list, tuple)):
        obj[key] = [obj[key]]

    if isinstance(obj[key], tuple):
        obj[key] = obj[key] + tuple(v)
    else:
        obj[key].extend(v)

    # Ensure unique values if requested
    if unique_values:
        if isinstance(obj[key], tuple):
            obj[key] = tuple(sorted(set(obj[key]), key=obj[key].index))
        else:
            obj[key] = sorted(set(obj[key]), key=obj[key].index)

    # Sort by given function if provided
    if sort_values_by:
        if isinstance(obj[key], tuple):
            obj[key] = tuple(sorted(obj[key], key=sort_values_by, reverse=reverse))
        else:
            obj[key].sort(key=sort_values_by, reverse=reverse)

    return obj


def dict_filter_value(obj: Dict[Any, Any], f) -> Any:
    """
    Results is a filtered dict by f func result

    @param obj: is a dict
    @param f: function filter
    @return:
    """
    inv_dict = invert_dict(obj)
    filter_val = f(inv_dict)
    res = inv_dict[filter_val]
    if isinstance(res, list):
        return dict(map(lambda x: (x, filter_val), res))
    return {res: filter_val}


def dict_max_value(obj: Dict[Any, Any]) -> Any:
    """
    Results is a filtered dict by max value

    >>> import cereja as cj
    >>> cj.dict_max_value({'oi': 10, 'aqui': 20, 'sei': 20})
    {'aqui': 20, 'sei': 20}

    @param obj: is a dict
    @return: dict filtered
    """
    return dict_filter_value(obj, max)


def dict_min_value(obj: Dict[Any, Any]) -> Any:
    """
    Results is a filtered dict by min value

    >>> import cereja as cj
    >>> cj.dict_min_value({'oi': 10, 'aqui': 20, 'sei': 20})
    {'oi': 10}

    @param obj: is a dict
    @return: dict filtered
    """
    return dict_filter_value(obj, min)


def get_zero_mask(number: int, max_len: int = 3) -> str:
    """
    Returns string of numbers formated with zero mask
    eg.
    >>> get_zero_mask(100, 4)
    '0100'
    >>> get_zero_mask(101, 4)
    '0101'
    >>> get_zero_mask(101, 4)
    '0101'
    """
    return f"%0.{max_len}d" % number


def get_batch_strides(data, kernel_size, strides=1, fill_=True, take_index=False):
    """
    Returns batches of fixed window size (kernel_size) with a given stride
    @param data: iterable
    @param kernel_size: window size
    @param strides: default is 1
    @param take_index: add number of index on items
    @param fill_: padding last batch if it needs
    """
    batches = []
    for index, item in enumerate(data):
        batches.append(item if not take_index else [index, item])
        if index % strides == 0 and len(batches) >= kernel_size:
            yield batches[:kernel_size]
            batches = batches[strides:]
    if len(batches):
        yield rescale_values(
                batches, granularity=kernel_size, filling="post"
        ) if fill_ else batches


def prune_values(values: Sequence, factor=2):
    assert is_indexable(values), TypeError("object is not subscriptable")
    if len(values) <= factor:
        return values
    w = round(len(values) / 2)
    k = int(round(w / factor))
    res = values[w - k: w + k]

    if len(res) == 0:
        return values[k]
    return res



"""
Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import threading
import time
from urllib import request as urllib_req
import json
import io
from ..config import PROXIES_URL

__all__ = ["HttpRequest", "HttpResponse"]

from urllib.error import HTTPError, URLError
from urllib.parse import urlparse

from cereja import Progress


class _Http:
    def __init__(self, url, data=None, headers=None, port=None):
        self.headers = headers or {}
        self._protocol, self._port, self._domains, self._endpoint = self.parse_url(
                url=url, port=port
        )
        self._data = data or None

    @staticmethod
    def is_url(val):
        try:
            result = urlparse(val)
            return all([result.scheme, result.netloc])
        except:
            return False

    @property
    def protocol(self):
        return self._protocol

    @property
    def port(self):
        return self._port

    @property
    def domains(self):
        return self._domains

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def data(self):
        return self._data

    @property
    def url(self):
        port = f":{self._port}" if self._port else self._port
        endpoint = f"/{self._endpoint}" if self._endpoint else self._endpoint
        return f"{self._protocol}://{self._domains}{port}{endpoint}"

    @property
    def content_type(self):
        return self.headers.get("Content-type")

    @property
    def content_length(self):
        return self.headers.get("Content-length")

    @classmethod
    def parse_url(cls, url: str, port=None):

        url = url.replace("://", ".")
        url = url.split("/", maxsplit=1)

        domains = url.pop(0).split(".")
        if ":" in domains[-1]:
            domain, port = domains[-1].split(":")
            domains[-1] = domain

        protocol = domains.pop(0) if domains[0].startswith("http") else "https"

        if not port:
            port = ""

        domains = ".".join(domains)
        endpoint = "/".join(url) if url else ""

        return protocol, port, domains, endpoint


class HttpResponse:
    CHUNK_SIZE = 1024 * 1024 * 1  # 1MB

    def __init__(self, request: "HttpRequest", save_on_path=None, timeout=None):
        self._finished = False
        self._timeout = timeout
        self._code = None
        self._status = None
        self._data = None
        self._headers = {}
        self._request = request
        self._save_on_path = save_on_path
        self._total_completed = "0.0%"
        self._th_request = threading.Thread(target=self.__request)
        self._th_request.start()
        while not self.headers and not self._finished:
            time.sleep(0.1)  # await headers

    def __request(self):
        # TODO: send to HttpRequest ?
        try:
            with urllib_req.urlopen(
                    self._request.urllib_req, timeout=self._timeout
            ) as req_file:
                self._code = req_file.status
                self._status = req_file.reason
                if hasattr(req_file, "headers"):
                    self._headers = dict(req_file.headers.items())
                if self.CHUNK_SIZE * 3 > (req_file.length or 0):
                    self._data = req_file.read()
                else:
                    with Progress(
                            name=f"Fetching data",
                            max_value=int(req_file.getheader("Content-Length")),
                            states=("download", "time"),
                    ) as prog:

                        with (
                                open(self._save_on_path, "wb")
                                if self._save_on_path
                                else io.BytesIO()
                        ) as f:
                            total_downloaded = 0
                            while True:
                                chunk = req_file.read(self.CHUNK_SIZE)
                                if not chunk:
                                    break
                                f.write(chunk)
                                total_downloaded += len(chunk)
                                prog.show_progress(total_downloaded)
                                self._total_completed = prog.total_completed

                            self._data = f if self._save_on_path else f.getvalue()
        except HTTPError as err:
            self._data = err.read()
            self._code = err.code
            self._status = err.reason
            if hasattr(err, "headers"):
                self._headers = dict(err.headers.items())
        except URLError as err:
            msg = f"{err.reason}: {self._request.url}"
            self._finished = True
            raise URLError(msg)
        self._finished = True

    def __repr__(self):
        if not self._finished:
            return f"<HttpResponse: Fetching data {self._total_completed}>"
        return f"<HttpResponse: code={self.code}, status={self._status}>"

    @property
    def content_type(self):
        if not self._finished:
            self._th_request.join()
        return self._headers.get("Content-Type")

    @property
    def headers(self):
        if not self._finished:
            self._th_request.join()
        return self._headers

    @property
    def request(self):
        if not self._finished:
            self._th_request.join()
        return self._request

    @property
    def success(self):
        if not self._finished:
            self._th_request.join()
        return self._code == 200

    @property
    def code(self):
        if not self._finished:
            self._th_request.join()
        return self._code

    @property
    def status(self):
        if not self._finished:
            self._th_request.join()
        return self._status

    def json(self):
        if not self._finished:
            self._th_request.join()
        return json.loads(self._data)

    @property
    def data(self):
        self._th_request.join()  # await for request
        if self.content_type == "application/json":
            if not self._data:
                return {}
            return self.json()
        return self._data


class HttpRequest(_Http):
    PROXIES = {}

    def __init__(self, method, url, *args, **kwargs):
        super().__init__(url=url, *args, **kwargs)
        if isinstance(self.data, dict):
            self.headers.update({"Content-type": "application/json"})
        self._count = 0
        self._method = method
        self._req = None

    def __repr__(self):
        return f"Request(url={self.url}, method={self._method})"

    @classmethod
    def get_proxies_list(cls):
        if cls.PROXIES:
            print("já tem proxies")
            return cls.PROXIES
        try:
            cls.PROXIES = json.loads(cls("GET", PROXIES_URL).send_request().data)
        except:
            pass
        return cls.PROXIES

    def send_request(self, save_on=None, timeout=None, **kwargs):
        if "data" in kwargs:
            self._data = self.parser(kwargs.pop("data"))

        if "headers" in kwargs:
            headers = kwargs.pop("headers")
            assert isinstance(headers, dict), TypeError(
                    "Headers type is invalid. send a dict"
            )
            self.headers.update(headers)
        self._count += 1
        return HttpResponse(request=self, save_on_path=save_on, timeout=timeout)

    @property
    def urllib_req(self):
        return urllib_req.Request(
                url=self.url,
                data=self.parser(self.data),
                headers=self.headers,
                method=self._method,
        )

    @property
    def total_request(self):
        return self._count

    @property
    def data(self):
        return self._data

    @classmethod
    def parser(cls, data) -> bytes:
        if isinstance(data, dict):
            return json.dumps(data).encode()

        return str(data).encode() if data else b""

    @classmethod
    def build_and_send(cls, method, url, data=None, port=None, headers=None, **kwargs):
        return cls(
                method=method, url=url, data=data, port=port, headers=headers
        ).send_request(**kwargs)



from collections import OrderedDict as _OrderedDict
import random

from cereja.utils import invert_dict, obj_repr

__all__ = ["CJOrderedDict", "CJDict", "CJMeta", "ParallelProcess"]


class CJMeta(type):
    __attr_limit = 5

    def __new__(mcs, name, bases, dct):
        x = super().__new__(mcs, name, bases, dct)
        x.__repr__ = lambda self: obj_repr(self, attr_limit=CJMeta.__attr_limit)
        return x


class CJOrderedDict(_OrderedDict):
    @property
    def first(self):
        k = next(iter(self))
        return k, self[k]


class CJDict(dict, metaclass=CJMeta):
    """
    Builtin dict class extension.

    What's new?

    methods:
    - item: get a item

    """

    __black_attr_list = dir(dict)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __invert__(self):
        """
        Invert dict values to key
        """
        return CJDict(invert_dict(self))

    def __getattr__(self, item):
        if item in self and item not in self.__black_attr_list:
            return self[item]
        return object.__getattribute__(self, item)

    def item(self, random_=True):
        if random_ is True:
            return random.choice(tuple(self.items()))
        k, v = self.popitem()
        self.update((k, v))
        return k, v


class DictOfList(CJDict):
    def add(self, key, value):
        if key not in self:
            self[key] = []
        self[key].append(value)

    def __setitem__(self, key, value):
        assert isinstance(value, list), "Send a list object"
        super(DictOfList, self).__setitem__(key, value)


import cereja as cj


class ParallelProcess:
    _template = """
from multiprocessing import Pool
import cereja as cj

JOB_DIR = cj.Path('{path}')

globals().update(cj.FileIO.load(JOB_DIR.join('global_scope.pkl')).data)

{func_code}

if __name__ == '__main__':
    with Pool({n_proc}) as p:
        sequence = cj.FileIO.load(JOB_DIR.join('sequence.pkl')).data
        result = p.map({func_name}, sequence)
        cj.FileIO.create(JOB_DIR.join('result.pkl'), result).save(exist_ok=True)
    """

    def __init__(self, function, global_scope: dict = None, n_proc=8):
        self.function = function
        self.global_scope = global_scope
        self.n_proc = n_proc

    def run(self, sequence):
        import sys

        with cj.TempDir() as temp_dir:
            func_code = cj.Source(self.function)
            code = self._template.format(
                    path=temp_dir.path,
                    func_code=func_code.source_code,
                    func_name=func_code.name,
                    n_proc=self.n_proc,
            )
            global_scope = (
                self.global_scope if isinstance(self.global_scope, dict) else {}
            )

            code_path = temp_dir.path.join("code.py")
            cj.FileIO.create(code_path, code).save()
            cj.FileIO.create(
                    temp_dir.path.join("global_scope.pkl"), global_scope
            ).save()
            cj.FileIO.create(temp_dir.path.join("sequence.pkl"), sequence).save()
            cj.run_on_terminal(f"{sys.executable} {code_path.path}")
            result = cj.FileIO.load(temp_dir.path.join("result.pkl")).data

        return result



"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import logging
from abc import abstractmethod, ABCMeta
import os

__all__ = ["BasicConfig", "BASE_DIR", "PYTHON_VERSION"]

# using by utils.module_references
_exclude = ["console_logger", "cj_modules_dotted_path"]

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

PYTHON_VERSION = sys.version_info[:3]


class BasicConfig(metaclass=ABCMeta):
    def __init__(self, hook=None, **kwargs):
        self._fields = ()
        self._listen = hook
        self.set_config(**kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.get()})"

    def __setattr__(self, key, value):
        if hasattr(self, "_locked"):
            if self._locked is True:
                er = "This is a critical config. You can use <set_config> method."
                raise ValueError(er)
        super().__setattr__(key, value)

    def __unlock(self):
        super().__setattr__("_locked", False)

    def __lock(self):
        self._fields = tuple(self.get().keys())
        super().__setattr__("_locked", True)

    def set_config(self, **kwargs) -> None:
        self._before(kwargs)
        self.__unlock()
        for k, i in kwargs.items():
            self.__setattr__(k, i)
        self.__lock()
        if self._listen is not None:
            self._listen()

    @abstractmethod
    def _before(self, new_config: dict):
        for k in new_config:
            if k not in self._fields and self._fields:
                raise ValueError(f"<{k}> not in config attrs {self._fields}")

    def get(self, item=None):
        config = {}
        if item is None:
            for attr_ in dir(self):
                if attr_.startswith("_"):
                    continue
                obj = self.__getattribute__(attr_)
                if callable(obj):
                    continue
                config[attr_] = obj
            return config
        if hasattr(self, item):
            obj = self.__getattribute__(item)
            if not callable(obj):
                return obj
        return {}


# Used to add the functions of each module at the root
cj_modules_dotted_path = [
    "cereja.array",
    "cereja.conf",
    "cereja.decorators",
    "cereja.path",
    "cereja.utils",
    "cereja.concurrently",
    "cereja.display",
    "cereja.filetools",
    "cereja.unicode",
    "cereja.datatools.split_data",
    "cereja.datatools.pln",
    "cereja.datatools.data",
]

console_logger = logging.StreamHandler(sys.stdout)
logging.basicConfig(handlers=(console_logger,), level=logging.WARNING)



"""
Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ._http import HttpRequest, HttpResponse

__all__ = [
    "is_url",
    "download",
    "get_proxies_list",
    "post",
    "get",
    "put",
    "head",
    "delete",
    "connect",
    "options",
    "trace",
    "patch",
]


def download(url, save_on=None, timeout=None, **kwargs) -> HttpResponse:
    """
    The GET method requests a representation of the specified resource. Requests using GET should only retrieve data.
    """
    return HttpRequest.build_and_send(
            method="GET", url=url, save_on=save_on, timeout=timeout, **kwargs
    )


def is_url(url):
    """
    Check if the url is valid.

    :param url: The url to check.
    :return: True if the url is valid, False otherwise.
    """
    return HttpRequest.is_url(url)


def get_proxies_list():
    return HttpRequest.get_proxies_list()


def post(url, data=None, headers=None, **kwargs) -> HttpResponse:
    """
    The POST method is used to submit an entity to the specified resource, often causing a change in state or side
    effects on the server.

    """
    return HttpRequest.build_and_send(
            method="POST", url=url, data=data, headers=headers, **kwargs
    )


def get(url, data=None, headers=None, **kwargs) -> HttpResponse:
    """
    The GET method requests a representation of the specified resource. Requests using GET should only retrieve data.
    """
    return HttpRequest.build_and_send(
            method="GET", url=url, data=data, headers=headers, **kwargs
    )


def put(url, **kwargs) -> HttpResponse:
    """
    The PUT method replaces all current representations of the target resource with the request payload.
    """
    return HttpRequest.build_and_send(method="PUT", url=url, **kwargs)


def head(url, **kwargs) -> HttpResponse:
    """
    The HEAD method asks for a response identical to that of a GET request, but without the response body.
    """
    return HttpRequest.build_and_send(method="HEAD", url=url, **kwargs)


def delete(url, **kwargs) -> HttpResponse:
    """
    The DELETE method deletes the specified resource.
    """
    return HttpRequest.build_and_send(method="DELETE", url=url, **kwargs)


def connect(url, **kwargs) -> HttpResponse:
    """
    The CONNECT method establishes a tunnel to the server identified by the target resource.
    """
    return HttpRequest.build_and_send(method="CONNECT", url=url, **kwargs)


def options(url, **kwargs) -> HttpResponse:
    """
    The OPTIONS method is used to describe the communication options for the target resource.
    """
    return HttpRequest.build_and_send(method="OPTIONS", url=url, **kwargs)


def trace(url, **kwargs) -> HttpResponse:
    """
    The TRACE method performs a message loop-back test along the path to the target resource.
    """
    return HttpRequest.build_and_send(method="TRACE", url=url, **kwargs)


def patch(url, **kwargs) -> HttpResponse:
    """
    The PATCH method is used to apply partial modifications to a resource.
    """
    return HttpRequest.build_and_send(method="PATCH", url=url, **kwargs)



import tempfile
import unittest
from abc import abstractmethod

from cereja.system import Path
from cereja.file import FileIO
from cereja.display import console


class FileIOTest(unittest.TestCase):
    data = None
    load_kwargs = {}

    @property
    def name(self):
        return self.__class__.__name__

    @abstractmethod
    def create(self):
        file = FileIO.create("test.tka", data=self.data)  # generic
        return file

    def setUp(self) -> None:
        self.file = self.create()

    def test_sanity(self):
        console.log(f"Testing {self.name}")
        with tempfile.TemporaryDirectory() as tempdir:
            self.file.set_path(Path(tempdir).join(self.file.name))
            data_before_save = self.file.data
            self.file.save(exist_ok=True)
            file = FileIO.load(self.file.path, **self.load_kwargs)
            self.assertTrue(file.path.exists, msg="File don't exist.")
            self.assertEqual(file.data, data_before_save, msg="Data corrupted")
            file.delete()
            self.assertFalse(file.path.exists)


class FileIOTxtTest(FileIOTest):
    data = ["first line", "second line", "third line"]

    def create(self):
        file = FileIO.txt("test.txt", data=self.data, creation_mode=True)
        return file

    def test_history(self):
        for line in self.file:
            pass

        self.assertEqual(self.file[0], "first line")
        self.assertEqual(self.file[:3], ["first line", "second line", "third line"])
        # Insert Data
        self.file.add("other line", line=0)
        self.file.add("other line2", line=0)
        self.assertEqual(
            self.file.data,
            ["other line2", "other line", "first line", "second line", "third line"],
        )
        # it is allowed to use index assignment
        self.file[0] = "other line"

        self.file.add("test")  # append
        self.assertEqual(
            self.file.data,
            [
                "other line",
                "other line",
                "first line",
                "second line",
                "third line",
                "test",
            ],
        )

        # Data Recovery
        self.file.undo()  # You selected amendment 4
        self.file.undo()  # You selected amendment 3
        self.assertEqual(
            self.file.data,
            ["other line2", "other line", "first line", "second line", "third line"],
        )
        self.file.redo()  # You selected amendment 4
        self.assertEqual(
            self.file.data,
            ["other line", "other line", "first line", "second line", "third line"],
        )

    def test_eval(self):
        with tempfile.TemporaryDirectory() as tempdir:
            eval_data = [1, 2, 3, "hi", ["oi"]]
            file = FileIO.create(Path(tempdir).join(self.file.name), eval_data)
            self.assertEqual(file.data, eval_data)
            file.save()
            file = FileIO.load(file.path)
            self.assertEqual(file.data, eval_data)


class FileIOJsonTest(FileIOTest):
    data = {"key": "value", "key2": "value2", "key3": "value3"}

    def create(self):
        return FileIO.json("test.json", data=self.data, creation_mode=True)

    def test_commons_operations(self):
        self.assertEqual(
            self.file.data, {"key": "value", "key2": "value2", "key3": "value3"}
        )

        # Iterable
        for key, value in self.file.items():
            pass

        self.assertEqual(self.file["key"], "value")

        # Insert Data
        self.file["key4"] = "value4"
        self.assertEqual(
            self.file.data,
            {"key": "value", "key2": "value2", "key3": "value3", "key4": "value4"},
        )

    def test_sample_items(self):
        self.assertEqual(self.file.sample_items(), {"key": "value"})
        self.assertEqual(
            self.file.sample_items(k=3),
            {"key": "value", "key2": "value2", "key3": "value3"},
        )
        self.assertRaises(AssertionError, self.file.sample_items, -1)
        self.assertRaises(AssertionError, self.file.sample_items, 0.5)


class FileCsvTest(FileIOTest):
    data = [[1, 2, 3]]
    cols = ["col1", "col2", "col3"]
    load_kwargs = {"has_col": True}

    def create(self):
        return FileIO.csv(
            "test.csv", data=self.data, creation_mode=True, cols=self.cols
        )

    def test_commons_operations(self):
        self.assertEqual(self.file.data, [[1, 2, 3]])
        self.file.add([1, 2], fill_with=0)
        self.assertEqual(self.file.data, [[1, 2, 3], [1, 2, 0]])

        # convert to dict
        self.assertEqual(
            self.file.to_dict(), {"col1": (1, 1), "col2": (2, 2), "col3": (3, 0)}
        )

        # or get generation row by row with col
        self.assertEqual(
            list(self.file.rows),
            [{"col1": 1, "col2": 2, "col3": 3}, {"col1": 1, "col2": 2, "col3": 0}],
        )

        # Iterable
        for row in self.file:
            pass

        # indexing col values
        self.assertEqual(self.file["col1"], (1, 1))
        # or use index and get a row
        self.assertEqual(self.file[0], [1, 2, 3])
        self.assertEqual(self.file.flatten(), [1, 2, 3, 1, 2, 0])


class FilePyObjTest(FileIOTest):
    data = [[1, 2, 3]]

    def create(self):
        return FileIO.create("test.pkl", data=self.data)

    def test_commons_operations(self):
        self.assertEqual(self.file.data, [[1, 2, 3]])


if __name__ == "__main__":
    unittest.main()



"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
from typing import AnyStr, Sequence, Union, List, Tuple
from unicodedata import normalize as _normalize

from ..config import (
    ENG_CONTRACTIONS,
    PUNCTUATION,
    VALID_LANGUAGE_CHAR,
    LANGUAGES,
    STOP_WORDS,
    NUMBER_WORDS
)

_NORMALIZE_VALUES = "".join(PUNCTUATION)


def split_by_punct(text: str, punct: str = '.!?') -> list:
    """
    split text by periods.

    :param text: any string
    :param punct: string of chars
    :return:
    """
    assert isinstance(punct, str), 'Send punct send the strings in string format: "!.?"'
    # Regular expression to split text by periods
    pattern = r'(?<=[%s])\s+' % punct
    texts = re.split(pattern, text)
    return texts


def convert_spelled_to_number(text, lang='pt'):
    words = NUMBER_WORDS.get(lang.lower())
    if words is None:
        raise ValueError("Unsupported language")

    text = text.replace(',', '')
    text = text.replace(' e ', ' ')
    text = text.replace('-', ' ')

    parts = text.split()
    result = []
    sum_values = []
    for n, word in enumerate(parts, start=1):
        if word in words:
            sum_values.append(words[word])
            if n == len(parts):
                result.append(str(sum(sum_values)))
            continue
        if len(sum_values):
            result.append(str(sum(sum_values)))
            sum_values = []
        result.append(word)
    return ' '.join(result)


def remove_delimited_text(text, delimiters: Sequence[Tuple[AnyStr, AnyStr]] = (("*", '*'), ("(", ")"), ("[", "]"))):
    """
    Remove words or phrases delimited by bullets from the provided text.

    Arguments:
      text(str): The text to be processed.
      delimiters (Sequence[Tuple[AnyStr, AnyStr]], optional): A sequence of tuples that specify the start and end markers of the delimited words or phrases.
          The default is (("*", '*'), ("(", ")"), ("[", "]")).

    Returns:
      str: The rendered text, without the delimited words or phrases.

    Example:
      >>> text = "Hello world of *options*! [shows options] How are you (walks out and says)?"
      >>> clear_text = remove_delimited_text(text)
      >>> print(clear_text)
      "Hello world! How are you?"
    """
    pattern = '|'.join(f'({re.escape(start)}{r".*?"}{re.escape(end)})' for start, end in delimiters)
    text = re.sub(pattern, "", text)
    return text


def separate(
        text: AnyStr, sep: Union[AnyStr, Sequence[AnyStr]] = PUNCTUATION, between_char=False
) -> str:
    """
    Creating a space between an element in sep values
    PUNCTUATION = {'?', '!', ',', '.'}
    e.g:

    >>> separate('how are you?', sep='?')
    'how are you ?'
    >>> separate('how are you,man?', sep=('?',','), between_char=True)
    'how are you , man ?'

    :param text: Any text
    :param sep: values that will be separated. Accepted types (str, list and tuple)
    :param between_char: if True the sep values that's on between chars it will be separated.
                         eg. "it's" -> "it ' s"
    :return: processed text
    """
    if isinstance(sep, str):
        sep = (sep,)
    assert isinstance(
            sep, (tuple, list)
    ), f"{type(sep)} is not acceptable. Only (str, list and tuple) types"
    assert isinstance(text, str), f"{type(text)} is not acceptable. Only str type."
    new_text = []
    for word in text.split():
        for i in sep:
            if (word.startswith(i) or word.endswith(i)) or between_char is True:
                word = f" {i} ".join(word.split(i)).strip()
        new_text += word.split()
    return " ".join(new_text)


def accent_remove(text: AnyStr) -> AnyStr:
    return _normalize("NFD", text).encode("ascii", "ignore").decode("UTF-8")


def replace_english_contractions(sentence: AnyStr) -> str:
    """
    Returns phrase without contractions
    """
    replaced_sentence = []
    for word in sentence.split(" "):
        if word.lower() in ENG_CONTRACTIONS:
            # Todo: create function that chooses the appropriate value
            word = ENG_CONTRACTIONS.get(word.lower()).split(",")[0]
        replaced_sentence.append(word)
    return " ".join(replaced_sentence)


def remove_extra_chars(sentence: str) -> str:
    """
    Replaces unnecessary blanks and punctuation
    analyzed values -> ['?', '!', ',', '.', ' ']
    e.g
        'i   like it!!!' --> 'i like it!'
    """
    sentence = re.sub(f"([{_NORMALIZE_VALUES}])+", r"\1", sentence)
    return re.sub("[\t\n\r\x0b\x0c ]+", " ", sentence).strip()


def remove_non_language_elements(sentence: str) -> str:
    for invalid in set(sentence).difference(VALID_LANGUAGE_CHAR):
        sentence = sentence.replace(invalid, "")
    return sentence.strip()


def remove_punctuation(sentence: str, punctuation: str = None):
    """
    Default Punctuation -> [',', '!', '#', '$', '%', "'", '*', '+', '-', '.', '/', '?', '@', '\\', '^', '_', '~']
    """
    punctuation = punctuation or "".join(PUNCTUATION)
    for x in punctuation:
        sentence = sentence.replace(x, "")
    return sentence.strip()


def remove_stop_words(
        sentence: str, stop_words: List[str] = None, language: str = "english"
):
    """
    Default Stop Words -> ['is', 'are', 'am', 'at', 'a', 'an', 'of', 'the']
    """
    language = language.lower()
    if language not in LANGUAGES:
        print(f'Idioma "{language}" não encontrado. Idiomas disponíveis: {LANGUAGES}')

    stop_words = stop_words if stop_words else STOP_WORDS.get(language, None)
    sentence = (
        " ".join([word for word in sentence.split() if word.lower() not in stop_words])
        if stop_words
        else sentence
    )
    return sentence.strip()


def preprocess(
        text: str,
        is_destructive: bool = False,
        is_lower: bool = None,
        is_remove_accent: bool = None,
        is_remove_punctuation: bool = None,
):
    """
    Run pack of preprocess functions

    non-destructive functions: remove_extra_chars, remove_non_language_elements and separate
    destructive functions: lower, accent_remove and remove_punctuation

    @param text: any string data.
    @param is_destructive: is a bool value, if True exec non-destructive and destructive functions
    @param is_lower: bool
    @param is_remove_accent: bool
    @param is_remove_punctuation: bool
    @return: preprocessed text
    """
    if is_destructive or is_lower:
        text = text.lower()
    text = remove_extra_chars(text)
    text = remove_non_language_elements(text)
    if is_destructive or is_remove_accent:
        text = accent_remove(text)

    if is_destructive or is_remove_punctuation:
        text = remove_punctuation(text)
    else:
        text = separate(text)
    return text



"""
Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import io
import os
import threading
import time
import sys
import random
from typing import Tuple, Callable
from abc import ABCMeta as ABC
from typing import List
from abc import ABCMeta, abstractmethod
from typing import Sequence, Any, Union, AnyStr

from cereja import has_length
from cereja.utils import is_iterable
from cereja.config.cj_types import Number
from cereja.system.unicode import Unicode
from cereja.utils import fill, time_format, get_instances_of, import_string
from cereja.mathtools import proportional, estimate, percent

__all__ = ["Progress", "State", "console"]

_exclude = [
    "_Stdout",
    "ConsoleBase",
    "BaseProgress",
    "ProgressLoading",
    "ProgressBar",
    "ProgressLoadingSequence",
    "State",
]

_include = ["console", "Progress", "StateBase"]

try:
    _LOGIN_NAME = os.getlogin()
except OSError:
    _LOGIN_NAME = "Cereja"

_SYS_EXCEPT_HOOK_ORIGINAL = sys.excepthook

try:
    # noinspection PyUnresolvedReferences
    IP = get_ipython()
    JUPYTER = True
except NameError:
    JUPYTER = False
    IP = None

_STDOUT_ORIGINAL_COPY = sys.stdout
_STDERR_ORIGINAL_COPY = sys.stderr


class _Stdout:
    __user_msg = []
    _stdout_original = _STDOUT_ORIGINAL_COPY
    _stderr_original = _STDERR_ORIGINAL_COPY

    def __init__(self, console_):
        self.persisting = False
        self.th_console = None
        self.last_console_msg = ""
        self.use_th_console = False
        self._stdout_buffer = io.StringIO()
        self._stderr_buffer = io.StringIO()
        self.console = console_

    @classmethod
    def die_threads(cls, *args, **kwargs):
        prog = import_string("cereja.display.Progress")
        for i in get_instances_of(prog):
            i.hook_error()
        if kwargs.get("is_jupyter") is None:
            _SYS_EXCEPT_HOOK_ORIGINAL(*args, **kwargs)

    @classmethod
    def custom_exc(cls, shell, etype, evalue, tb, tb_offset=None):
        shell.showtraceback((etype, evalue, tb), tb_offset=tb_offset)
        cls.die_threads(is_jupyter=True)

    def set_message(self, msg):
        self.__user_msg = msg

    def _has_user_msg(self):
        if self._stdout_buffer.getvalue():
            return True
        else:
            return False

    @property
    def stdout_buffer_values(self):
        vals = self._stdout_buffer.getvalue()
        self._stdout_buffer.seek(0)
        self._stdout_buffer.truncate()
        return vals

    @property
    def stderr_buffer_values(self):
        vals = self._stderr_buffer.getvalue()
        self._stderr_buffer.seek(0)
        self._stderr_buffer.truncate()
        return vals

    def write_user_msg(self):
        stdout, stderr = self.stdout_buffer_values, self.stderr_buffer_values
        if stdout and not (stdout in ["\n", "\r\n", "\n"]):
            values = []
            for value in stdout.splitlines():
                value = f"{self.console.parse(value, title='Sys[out]')}"
                values.append(value)
            value = "\n".join(values)
            value = f"\r{value}\n{self.last_console_msg}"
            self._write(value)

        if stderr and not (stderr in ["\n", "\r\n", "\n"]):
            unicode_err = "\U0000274C"
            prefix = self.console.template_format(
                    f"{{red}}{unicode_err} Error:{{endred}}"
            )
            msg_err_prefix = f"{self.console.parse(prefix, title='Sys[err]')}"
            msg_err = self.console.format(stderr, color="red")
            msg_err = f"\r{msg_err_prefix}\n{msg_err}\n{self.last_console_msg}"
            self._write(msg_err)

    def _write_user_msg_loop(self):
        while self.use_th_console:
            self.write_user_msg()
            time.sleep(0.1)
        if self._has_user_msg():
            self.write_user_msg()

    def _write(self, msg: str):
        try:

            if msg.strip(" ") == "":
                return
            self._stdout_original.write(msg)
            self._stdout_original.flush()
        except (UnicodeError, UnicodeEncodeError):
            msg = self.console.translate_non_bmp(msg)
            self._stdout_original.write(msg.encode("ascii", errors="replace").decode())
            self._stdout_original.flush()

    def cj_msg(self, msg: str, line_sep=None, replace_last=False):
        self.last_console_msg = msg
        if replace_last:
            msg = f"\r{msg}"
        self._write(msg)
        if line_sep is not None:
            self._write(line_sep)

    def __del__(self):
        self.restore_sys_module_state()

    def write_error(self, msg, **ars):
        msg = self.console.format(msg, color="red")
        self._stdout_buffer.write(msg)

    def flush(self):
        self._stdout_original.flush()

    def persist(self):
        if not self.persisting:
            self.use_th_console = True
            self.th_console = threading.Thread(
                    name="Console", target=self._write_user_msg_loop
            )
            self.th_console.start()
            sys.stdout = self._stdout_buffer
            sys.stderr = self._stderr_buffer
            self.persisting = True

    def disable(self):
        if self.use_th_console:
            self.use_th_console = False
            self.th_console.join()
            self.persisting = False
            self.console.set_prefix(_LOGIN_NAME)
        self.restore_sys_module_state()

    def restore_sys_module_state(self):
        if JUPYTER:
            IP.restore_sys_module_state()
            return
        if not isinstance(self._stdout_original, io.StringIO):
            sys.stdout = _STDOUT_ORIGINAL_COPY
            sys.stderr = _STDERR_ORIGINAL_COPY


class _ConsoleBase(metaclass=ABC):
    NON_BMP_MAP = dict.fromkeys(range(0x10000, sys.maxunicode + 1), 0xFFFD)
    DONE_UNICODE = "\U00002705"
    ERROR_UNICODE = "\U0000274C"
    CHERRY_UNICODE = "\U0001F352"
    RIGHT_POINTER = "\U000000bb"
    __CL_DEFAULT = "\033[0;0m"
    CL_BLACK = "\033[30m"
    CL_RED = "\033[31m"
    CL_GREEN = "\033[38;5;2m"
    CL_YELLOW = "\033[33m"
    CL_BLUE = "\033[34m"
    CL_MAGENTA = "\033[35m"
    CL_CYAN = "\033[36m"
    CL_WHITE = "\033[37m"
    CL_UNDERLINE = "\033[4m"
    __line_sep_list = ["\r\n", "\n", "\r"]
    __os_line_sep = os.linesep

    __right_point = f"{CL_CYAN}{RIGHT_POINTER}{__CL_DEFAULT}"
    __msg_prefix = f"{CL_RED}{CHERRY_UNICODE}{CL_BLUE}"
    __color_map = {
        "black":   CL_BLACK,
        "red":     CL_RED,
        "green":   CL_GREEN,
        "yellow":  CL_YELLOW,
        "blue":    CL_BLUE,
        "magenta": CL_MAGENTA,
        "cyan":    CL_CYAN,
        "white":   CL_WHITE,
        "default": __CL_DEFAULT,
    }

    MAX_SPACE = 20
    MAX_BLOCKS = 5

    SPACE_BLOCKS = {""}
    _name = _LOGIN_NAME

    def __init__(self, color_text: str = "default"):
        self.__color_map = self._color_map  # get copy
        self.text_color = color_text
        self._stdout = _Stdout(self)

    @property
    def non_bmp_supported(self):
        from cereja import NON_BMP_SUPPORTED

        return NON_BMP_SUPPORTED

    @property
    def prefix_name(self):
        return self._name

    def set_prefix(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Please send string value.")
        self._name = value

    @property
    def _color_map(self):
        return self.__color_map.copy()

    @property
    def text_color(self):
        return self.__text_color

    @text_color.setter
    def text_color(self, color: str):
        color = (
            color.strip("CL_").lower()
            if color.upper().startswith("CL_")
            else color.lower()
        )
        if color not in self.__color_map:
            raise ValueError(f"Color not found.")
        self.__color_map["default"] = self.__color_map[color]
        self.__text_color = self.__color_map[color]

    def __bg(self, text_, color):
        return f"\33[48;5;{color}m{text_}{self.__text_color}"

    def _msg_prefix(self, title=None):
        if title is None:
            title = self.prefix_name
        return f"{self.__msg_prefix} {title} {self.__right_point}"

    def translate_non_bmp(self, msg: str):
        """
        This is important, as there may be an exception if the terminal does not support unicode bmp
        """
        if self.non_bmp_supported:
            return msg
        return msg.translate(self.NON_BMP_MAP)

    def _parse(self, msg, title=None, color: str = None, remove_line_sep=True):

        if color is None:
            color = "default"
        msg = str(msg)
        msg = self.format(msg, color)
        if remove_line_sep:
            msg = " ".join(msg.splitlines())
        prefix = self._msg_prefix(title)
        return f"{prefix} {msg} {self.__CL_DEFAULT}"

    def _write(self, msg, line_sep=None):
        self._stdout.cj_msg(msg, line_sep)

    def _normalize_format(self, s):
        for color_name in self.__color_map:
            s = s.replace("{end%s}" % color_name, self.__color_map["default"])
            s = s.replace("{%s}" % color_name, self.__color_map[color_name])
        s = s.replace("  ", " ") + self.__color_map["default"]
        return s

    def parse(self, msg, title=None):
        return self._parse(msg, title)

    def template_format(self, s):
        """
        You can make format with variables
        :param s:
        :return:
        """
        return self._normalize_format(s)

    def format(self, s: str, color: str):
        if color == "random":
            return self.random_color(s)
        if color not in self.__color_map:
            raise ValueError(
                    f"Color {repr(color)} not found. Choose an available color"
                    f" [red, green, yellow, blue, magenta and cyan]."
            )
        s = self._normalize_format(s)
        return f"{self._color_map[color]}{s}{self._color_map['default']}"

    def random_color(self, text: str):
        color = random.choice(list(self.__color_map))
        template_format = f'{{{color}}}{text}{{{"end" + color}}}'
        return self.template_format(template_format)

    def colorful_words(self, text: Union[List[str], str]) -> str:
        msg_error = "You need send string or list of string."
        if isinstance(text, str):
            text = text.split()
        elif isinstance(text, list):
            if not isinstance(next(iter(text)), str):
                self.error(msg_error)
        else:
            self.error(msg_error)
        result = []
        for word in text:
            result.append(self.random_color(word))
        return " ".join(result)

    def text_bg(self, text_, color):
        return self.__bg(text_, color)

    def error(self, msg: str):
        msg = f"Error: {msg}"
        msg = self.format(msg, color="red")
        self.__print(msg)

    def replace_last_msg(self, msg: str, end=None):
        msg = self._parse(msg, remove_line_sep=True)
        self._stdout.cj_msg(msg, replace_last=True, line_sep=end)

    def log(self, msg, title=None, color: str = "default", end: str = __os_line_sep):
        """
        Send message with default configuration and colors
        title is login name.

        :param msg: value you want to display
        :param title: You can custom title of message
        :param color: You can custom color of message
                colors available: black, red, green, yellow, blue, magenta, cyan and white.
        :param end: if you send None or '' remove line_sep
        """
        self.__print(msg=msg, title=title, color=color, end=end)

    def persist_on_runtime(self):
        self._stdout.persist()

    def __print(self, msg, title=None, color: str = None, end=__os_line_sep):
        remove_line_sep = True if end is None else False
        msg = self._parse(
                msg=msg, title=title, color=color, remove_line_sep=remove_line_sep
        )
        self._stdout.cj_msg(msg, end)

    def disable(self):
        self._stdout.disable()
        self._name = _LOGIN_NAME


console = _ConsoleBase()


class State(metaclass=ABCMeta):
    size = 10

    def __repr__(self):
        return f"{self.name} {self.done(100, 100, 100, 0, 100)}"

    def blanks(self, current_size):
        blanks = "  " * (self.size - current_size - 1)
        return blanks

    @property
    def name(self):
        return f"{self.__class__.__name__.replace('__State', '')} field"

    @abstractmethod
    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
    ) -> str:
        """
        This function is always being called.
        :return: You need to return a string with the status of your progress
        """
        pass

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
    ) -> str:
        return self.display(current_value, max_value, current_percent, time_it, n_times)

    def error(self, *args, **kwargs) -> str:
        return self.display(*args, **kwargs)


class _StateLoading(State):
    """
    lef_right_delimiter: str = String of length 2 e.g: '[]'
            Are the delimiters of the default_char
            The default value is '[]'.
            e.g: [=============] # left is '[' and right is ']'


    default_char: Single string you can send a unicode :)
            Value that will be added or replaced simulating animation.
    """

    sequence = (".", ".", ".")
    left_right_delimiter = "[]"
    default_char = "."
    size = 3

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        idx = n_times % self.size
        value = "".join(self.sequence[: idx + 1])
        filled = fill(value, self.size, with_=" ")
        l_delimiter, r_delimiter = self.left_right_delimiter
        return f"{l_delimiter}{filled}{r_delimiter}"

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        l_delimiter, r_delimiter = self.left_right_delimiter
        return f"{l_delimiter}{self.default_char * self.size}{r_delimiter}"


class _StateAwaiting(_StateLoading):
    def _clean(self, result):
        return result.strip(self.left_right_delimiter)

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        result = self._clean(
                super().display(
                        current_value=current_value,
                        max_value=max_value,
                        current_percent=current_percent,
                        time_it=time_it,
                        n_times=n_times,
                )
        )
        if kwargs.get("is_generator"):
            return f"{time_format(time_it)} - Total: {current_value} - Processing{result}"
        return f"{time_format(time_it)} - Awaiting{result}"

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        return f"Total Time: {time_format(time_it)}"


class _StateDownloadData(State):
    _size_map = {"B": 1.0e0, "KB": 1.0e3, "MB": 1.0e6, "GB": 1.0e9, "TB": 1.0e12}

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        current_value /= self._size_map["MB"]
        max_value /= self._size_map["MB"]
        per_second = current_value / time_it
        return f"{current_value:.2f}/{max_value:.2f} MB ({per_second:.2f} MB/s)"


class _StateBar(State):
    left_right_delimiter = "[]"
    arrow = ">"
    default_char = "="
    blank = " "
    size = 30

    def __init__(self):
        if console.non_bmp_supported:
            self.arrow = ""
            self.default_char = "▰"
            self.blank = "▱"

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> AnyStr:
        l_delimiter, r_delimiter = self.left_right_delimiter
        prop_max_size = int(proportional(current_value, max_value, self.size))
        blanks = self.blank * (self.size - prop_max_size - 1)
        body = console.template_format(
                f"{{green}}{self.default_char * prop_max_size}{self.arrow}{{endgreen}}"
        )
        # value = f"{current_value:0{len(str(max_value))}d}/{max_value}"
        return f"{l_delimiter}{body}{blanks}{r_delimiter}"

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        l_delimiter, r_delimiter = self.left_right_delimiter
        body = console.template_format(
                f"{{green}}{self.default_char * self.size}{{endgreen}}"
        )
        return f"{l_delimiter}{body}{r_delimiter}"


class _StatePercent(State):
    size = 8

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        return f"{current_percent:.2f}%"

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        return f"{100:.2f}%"


class _StateTime(State):
    __clock = Unicode("\U0001F55C")
    __max_sequence = 12
    size = 11

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        time_estimate = estimate(
                current_value - kwargs.get("started_on", 0),
                max_value - kwargs.get("started_on", 0),
                time_it,
        )
        idx = int(proportional(current_value, max_value, self.__max_sequence))
        t_format = f"{time_format(time_estimate)}"
        value = f"{self.__clock + idx} {time_format(time_it)}/{t_format}"
        return f"{value}{self.blanks(len(value))} estimated"

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: float,
            **kwargs,
    ) -> str:
        return f"{self.__clock} {time_format(time_it)} total"


class _StateValue(State):
    _update_size = False
    _mask = "%.d"

    def display(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        current_value = f"{current_value:0{len(str(max_value))}d}"
        return f"{current_value}/{max_value}"

    def done(
            self,
            current_value: Number,
            max_value: Number,
            current_percent: Number,
            time_it: Number,
            n_times: int,
            **kwargs,
    ) -> str:
        return f"{max_value}/{max_value}"


class Progress:
    """
    Percentage calculation is based on max_value (default is 100) and current_value:
    percent = (current_value / max_value) * 100

    :param name: Defines the name to be displayed in progress
    :param max_value: This number represents the maximum amount you want to achieve.
                      It is not a percentage, although it is purposely set to 100 by default.
    """

    __awaiting_state = _StateAwaiting()
    __done_unicode = Unicode("\U00002705")
    __err_unicode = Unicode("\U0000274C")
    __key_map = {
        "loading":  _StateLoading,
        "time":     _StateTime,
        "percent":  _StatePercent,
        "bar":      _StateBar,
        "value":    _StateValue,
        "download": _StateDownloadData,
    }
    _with_context = False
    __built = False
    _console = console
    _progresses = {}

    def __init__(
            self,
            sequence=None,
            name="Progress",
            max_value: int = 100,
            states=("value", "bar", "percent", "time"),
            custom_state_func=None,
            custom_state_name=None,
    ):
        self._is_generator = False
        self._n_times = 0
        self._name = name or "Progress"
        self._task_count = 0
        self._started = False
        self._awaiting_update = True
        self._iter_finaly = False
        self._show = False
        self._custom_state_value = (
            custom_state_func if isinstance(custom_state_func, Callable) else None
        )
        self._custom_state_name = (
            custom_state_name if custom_state_name else "Custom State"
        )
        self._started_time = None
        self._states = ()
        self.add_state(states)
        self._max_value = max_value
        self._current_value = 0
        self._th_root = self._create_progress_service()
        self._was_done = False
        self._err = False
        self.__built = True
        if sequence is not None:
            self.__call__(sequence)
        if len(Progress._progresses) >= 10:
            # TODO: This is a hack to avoid memory leaks.
            Progress._progresses.clear()
        Progress._progresses[id(self)] = self

    @property
    def name(self):
        return self._name

    def set_name(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Please send string.")
        self._name = value

    def _create_progress_service(self):
        return threading.Thread(
                name="awaiting", target=self._progress_service, daemon=True
        )

    def __repr__(self):
        state = self._states_view(self._max_value)
        progress_example_view = f"{state}"
        state_conf = f"{self.__class__.__name__}{self._parse_states()}"
        return f"{state_conf}\n{self._console.parse(progress_example_view, title='Example States View')}"

    def hook_error(self, *args, **kwargs):
        self._err = True
        if not self._with_context:
            self.stop()

    @property
    def completed(self):
        return self._was_done or self._err or self._iter_finaly

    @staticmethod
    def die_all():
        for progress in Progress._progresses.values():
            progress.stop()

    def _parse_states(self):
        return tuple(map(lambda stt: stt.__class__.__name__, self._states))

    def _get_done_state(self, **kwargs):
        if self._current_value == 0:
            result = [self.__awaiting_state.done(**kwargs)]
        else:
            result = list(map(lambda state: state.done(**kwargs), self._states))
        done_msg = f"Done! {self.__done_unicode}"
        done_msg = self._console.format(done_msg, "green")
        result.append(done_msg)
        return result

    def _get_state(self, **kwargs):
        return list(map(lambda state: state.display(**kwargs), self._states))

    @property
    def time_it(self):
        return time.time() - (self._started_time or time.time())

    @property
    def total_completed(self):
        return f"{self.percent_(self._current_value)}%"

    def _states_view(self, for_value: Number) -> str:
        """
        Return current state and bool
        bool:
        :param for_value:
        :return: current state and bool if is done else False
        """
        self._n_times += 1
        kwargs = {
            "current_value":   for_value,
            "max_value":       self._max_value,
            "current_percent": self.percent_(for_value),
            "time_it":         self.time_it,
            "n_times":         self._n_times,
        }
        extra_ = (
            f" - {self._custom_state_name}: {self._custom_state_value()}"
            if self._custom_state_value
            else ""
        )
        if self._was_done:
            return " - ".join(self._get_done_state(**kwargs)) + extra_
        else:
            _state_msg = " - ".join(self._get_state(**kwargs)) + extra_
            if self._err:
                error_msg = f"Interrupted!"
                error_msg = self._console.format(error_msg, "red")
                _state_msg = f"{_state_msg} {error_msg}"
            elif self._iter_finaly:
                finish_msg = f"Done! {self.__done_unicode}"
                finish_msg = self._console.format(finish_msg, "green")
                _state_msg = f"{_state_msg} {finish_msg}"
            return _state_msg

    def add_state(self, state: Union[State, Sequence[State]], idx=-1):
        self._filter_and_add_state(state, idx)

    def remove_state(self, idx):
        states = list(self._states)
        states.pop(idx)
        self._states = tuple(states)

    def _parse_state(self, state):
        if isinstance(state, str):
            state = self.__key_map.get(state)
        assert issubclass(state, State) or isinstance(
                state, State
        ), f"State `{state}` not found in valid values [{list(self.__key_map)}]"
        if not isinstance(state, State):
            return state()
        return state

    def _valid_states(self, states: Union[State, Sequence[State]]):
        if states is not None:
            if not is_iterable(states):
                states = (states,)
            return tuple(map(self._parse_state, states))
        return (None,)

    def _filter_and_add_state(self, state: Union[State, Sequence[State]], index_=-1):
        state = self._valid_states(state)
        filtered = tuple(filter(lambda stt: stt not in self._states, tuple(state)))
        if any(filtered):
            if index_ == -1:
                self._states += filtered
            else:
                states = list(self._states)
                for idx, new_state in enumerate(filtered):
                    states.insert(index_ + idx, new_state)
                self._states = tuple(states)
            if self.__built:
                self._console.log(f"Added new states! {filtered}")

    @property
    def states(self):
        return self._parse_states()

    def percent_(self, for_value: Number) -> Number:
        return percent(for_value, self._max_value)

    def update_max_value(self, max_value: int):
        """
        You can modify the progress to another value. It is not a percentage,
        although it is purposely set to 100 by default.

        :param max_value: This number represents the maximum amount you want to achieve.
        """
        if not isinstance(max_value, (int, float, complex)):
            raise Exception(f"Current value {max_value} isn't valid.")
        if max_value != self._max_value:
            self._max_value = max_value

    def _progress_service(self):
        last_value = self._current_value
        n_times = 0
        while self._started and not self._iter_finaly:
            if (self._awaiting_update and self._current_value != last_value) or (
                    not self._show and not self._was_done
            ):
                n_times += 1
                self._console.replace_last_msg(
                        self.__awaiting_state.display(
                                self._current_value, 0, 0, time_it=self.time_it, n_times=n_times, is_generator=self._is_generator
                        )
                )
                time.sleep(0.5)
            if not self._awaiting_update or self._show:
                self._show_progress(self._current_value)
            last_value = self._current_value
            time.sleep(0.01)

    def _show_progress(self, for_value=None):
        self._awaiting_update = False
        build_progress = self._states_view(for_value)
        self._console.replace_last_msg(
                build_progress, end="\n" if self._iter_finaly else None
        )

    def _update_value(self, value):
        self._awaiting_update = self._is_generator  # if is generator awaiting state is default.
        self._show = True
        self._current_value = value

    def show_progress(self, for_value, max_value=None):
        """
        Build progress by a value.

        :param for_value: Fraction of the "max_value" you want to achieve.
                              Remember that this value is not necessarily the percentage.
                              It is totally dependent on the "max_value"
        :param max_value: This number represents the maximum amount you want to achieve.
                          It is not a percentage, although it is purposely set to 100 by default.
        """
        if max_value is not None:
            self.update_max_value(max_value)
        self._update_value(for_value)

    def _reset(self):
        self._current_value = 0
        self._n_times = 0
        self._started_time = None

    def start(self):
        self._console.set_prefix(self._name)
        if self._started:
            return
        self._started_time = time.time()
        self._started = True
        try:
            self._th_root.start()
        except:
            self._th_root.join()
            self._th_root = self._create_progress_service()
            self._th_root.start()
        self._console.persist_on_runtime()
        self._n_times = 0

    def stop(self):
        if self._started:
            self._awaiting_update = False
            self._started = False
            self._th_root.join()
            self._console.disable()
        if id(self) in Progress._progresses:
            # TODO: verify if this works
            Progress._progresses.pop(id(self))

    def restart(self):
        self._reset()

    @classmethod
    def prog(
            cls,
            sequence: Sequence[Any],
            name: str = None,
            states=("value", "bar", "percent", "time"),
            custom_state_func=None,
            custom_state_name=None,
    ) -> "Progress":
        return cls(
                name=name,
                states=states,
                custom_state_func=custom_state_func,
                custom_state_name=custom_state_name,
        )(sequence)

    def __len__(self):
        return len(self._states)

    def __getitem__(self, slice_):
        if isinstance(slice_, tuple):
            if max(slice_) > len(self):
                raise IndexError(f"{max(slice_)} isn't in progress")
            return tuple(self._states[idx] for idx in slice_ if idx < len(self))
        if isinstance(slice_, str):
            key = self.__key_map[slice_]
            if key in self._states:
                slice_ = self._states.index(key)
            else:
                raise KeyError(f"Not exists {key}")
        return self._states[slice_]

    def __call__(self, sequence: Sequence, name=None) -> "Progress":
        if not is_iterable(sequence):
            raise ValueError("Send a sequence.")
        if has_length(sequence):
            self.update_max_value(len(sequence))
        else:
            self._is_generator = True
        self.sequence = sequence
        if name is not None:
            self._console.set_prefix(f"{self.name}({name})")
        if self._with_context and name is None:
            self._console.set_prefix(f"{self.name}(iter-{self._task_count})")
        self._started_time = time.time()
        self._task_count += 1
        return self

    def __next__(self):
        if not self._with_context:
            self.start()
        try:
            for n, obj in enumerate(self.sequence):
                self._update_value(n + 1)
                yield obj
        except:
            self._err = True
        finally:
            if self._is_generator:
                self.update_max_value(self._current_value)
            self._was_done = (self._current_value >= self._max_value) and not self._err
            if not self._with_context:
                self.stop()
            self._iter_finaly = True
            self._show_progress(self._current_value)

        self._console.set_prefix(self.name)
        self.sequence = ()

    def __iter__(self):
        return self.__next__()

    def __setitem__(self, key, value):
        value = self._valid_states(value)[0]
        if isinstance(value, State):
            value = value
            if isinstance(key, int):
                states_ = list(self._states)
                states_[key] = value
                self._states = tuple(states_)
        else:
            raise ValueError("Please send State object")

    def __enter__(self, *args, **kwargs):
        if hasattr(self, "sequence"):
            raise ChildProcessError("Dont use progress instance on with statement.")
        self._with_context = True
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_val, Exception) and not isinstance(
                exc_val, DeprecationWarning
        ):
            self._console.error(
                    f"{os.path.basename(exc_tb.tb_frame.f_code.co_filename)}:{exc_tb.tb_lineno}: {exc_val}"
            )
        self.stop()
        self._task_count = 0
        self._with_context = False


if IP:
    IP.set_custom_exc(
            (Exception, GeneratorExit, SystemExit, KeyboardInterrupt), _Stdout.custom_exc
    )
else:
    sys.excepthook = _Stdout.die_threads



import copy
import csv
import logging
import os
import re
import pickle
import tempfile
from abc import ABCMeta, abstractmethod, ABC
from io import BytesIO
from typing import (
    Any,
    List,
    Union,
    Type,
    TextIO,
    Iterable,
    Tuple,
    KeysView,
    ItemsView,
    ValuesView,
)
import json

from .._requests import request
from zipfile import ZipFile, ZIP_DEFLATED

from ..display import Progress
from ..config import DATA_UNIT_MAP
from ..system import Path, mkdir
from ..array import get_cols, flatten, get_shape
from ..utils import is_sequence, clipboard
from ..system import memory_of_this
from ..utils import string_to_literal, sample, fill

logger = logging.Logger(__name__)

__all__ = ["FileIO"]


class _IFileIO(metaclass=ABCMeta):
    _size_map = DATA_UNIT_MAP
    _dont_read = [".pyc"]
    _ignore_dir = [".git"]
    _date_format = "%Y-%m-%d %H:%M:%S"
    # need implement
    _is_byte = None
    _only_read = None
    _newline = True
    _ext_allowed = None
    _need_implement = {
        "_is_byte":     """
    Define the configuration of the file in the class {class_name}:
    e.g:
    class {class_name}:
        _is_byte: bool = True # or False
    """,
        "_only_read":   """Define the configuration of the file in the class {class_name}:
    e.g:
    class {class_name}:
        _only_read: bool = True # or False

    """,
        "_ext_allowed": """Define the configuration of the file in the class {class_name}:
                       e.g:
                       class {class_name}:
                           _ext_allowed: bool = '.txt' # or ''

                       """,
    }

    def __init__(self):
        self._built = False
        self._length = None
        self._last_update = None
        self._is_deleted = False
        self._current_change = 0
        self._max_history_length = 50
        self._change_history = []
        self._data = None

    @property
    @abstractmethod
    def data(self) -> Any:
        pass

    @property
    @abstractmethod
    def path(self) -> Path:
        pass

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def ext(self):
        pass

    @property
    @abstractmethod
    def name_without_ext(self):
        pass

    @property
    @abstractmethod
    def dir_name(self):
        pass

    @property
    @abstractmethod
    def dir_path(self):
        pass

    @property
    @abstractmethod
    def is_empty(self):
        pass

    @property
    @abstractmethod
    def updated_at(self):
        pass

    @property
    @abstractmethod
    def created_at(self):
        pass

    @property
    @abstractmethod
    def was_changed(self) -> bool:
        pass

    @property
    @abstractmethod
    def length(self):
        pass

    @property
    @abstractmethod
    def history(self):
        pass

    @property
    @abstractmethod
    def exists(self):
        pass

    @property
    @abstractmethod
    def last_access(self):
        pass

    @abstractmethod
    def _load(self, mode=None, **kwargs) -> Any:
        pass

    @abstractmethod
    def save(
            self,
            on_new_path: Union[str, Path] = None,
            exist_ok=False,
            overwrite=False,
            **kwargs,
    ):
        pass

    @abstractmethod
    def _parse_fp(self, fp: Union[TextIO, BytesIO]) -> Any:
        pass

    @abstractmethod
    def _save_fp(self, fp: Union[TextIO, BytesIO]) -> None:
        pass

    @abstractmethod
    def undo(self):
        pass

    @abstractmethod
    def redo(self):
        pass

    @abstractmethod
    def parse(self, data):
        pass

    @abstractmethod
    def add(self, data, **kwargs):
        pass

    @abstractmethod
    def size(self, unit: str = "KB"):
        pass

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def set_path(self, p: str):
        pass


class _FileIO(_IFileIO, ABC):
    def __init__(
            self, path_: Path, data=None, creation_mode=False, is_byte=None, **kwargs
    ):
        super().__init__()
        if is_byte is not None:
            self._is_byte = bool(is_byte)
        if not isinstance(path_, Path):
            path_ = Path(path_)
        self._creation_mode = creation_mode
        self._path = path_
        if self._ext_allowed:
            msg_assert_err = f'{path_.suffix} != {self._ext_allowed}. Force with support_ext = ".xyz" '
            assert self._path.suffix in self._ext_allowed, msg_assert_err
        self._ext_without_point = self._path.suffix.strip(".").upper()
        self._built = (
            True  # next setattr will be saved in history to be future recovery
        )
        if creation_mode:
            self._data = self.parse(data)
        else:
            self._data = self._load(**kwargs)

    def __init_subclass__(cls, **kwargs):
        for attr in cls._need_implement:
            if getattr(cls, attr) is None:
                raise NotImplementedError(
                        cls._need_implement.get(attr).format(class_name=cls.__name__)
                )

    def __repr__(self):
        return f"File{self.__class__.__name__.replace('_', '')}({self._path.name})"

    def __setattr__(self, key, value):
        object.__setattr__(self, key, copy.copy(value))
        if (
                hasattr(self, "_change_history")
                and key
                not in (
                "_current_change",
                "_max_history_length",
                "_change_history",
                "_built",
                "_last_update",
        )
                and self._built is True and not self._only_read
        ):
            self._set_change(
                    key, copy.copy(object.__getattribute__(self, key))
            )  # append last_value of attr

    def __getitem__(self, item):
        try:
            return self.data[item]
        except TypeError:
            raise TypeError(f"'{self.__class__.__name__}' object is not subscriptable")

    def __setitem__(self, key, value):
        # TODO: save only the index that has been changed
        try:
            self._data[key] = value
            self._set_change("_data", copy.copy(self._data))
        except TypeError:
            raise TypeError(f"'{self.__class__.__name__}' object is not subscriptable")

    def _set_change(self, key, value):
        self._change_history = self._change_history[: self._current_change + 1]
        if len(self._change_history) >= self._max_history_length:
            self._change_history.pop(0)
        self._change_history.append((key, value))
        self._current_change = len(self._change_history)

    def _select_change(self, index):
        try:

            key, value = self._change_history[self._current_change + index]
            object.__setattr__(self, key, copy.copy(value))
            self._current_change += index
            logger.warning(f"You selected amendment {self._current_change + 1}")
        except IndexError:
            logger.info("It's not possible")

    @property
    def line_sep(self):
        return os.linesep

    @property
    def was_changed(self) -> bool:
        """
        Detect if file is changed by another application
        if file not exists return False

        @return: bool
        """
        if not self.exists:
            return False
        return self._last_update is not None and self.updated_at != self._last_update

    @property
    def length(self):
        return len(self._data)

    @property
    def history(self):
        return self._change_history

    @property
    def updated_at(self):
        return self._path.updated_at

    @property
    def created_at(self):
        return self._path.created_at

    @property
    def last_access(self):
        return self._path.last_access

    @property
    def name(self):
        return self._path.name

    @property
    def ext(self):
        return self._path.suffix

    @property
    def name_without_ext(self):
        return self._path.stem

    @property
    def data(self):
        return copy.copy(self._data)

    @property
    def path(self):
        return self._path

    @property
    def dir_name(self):
        return self._path.parent_name

    @property
    def dir_path(self):
        return self._path.parent

    @property
    def is_empty(self):
        return not bool(self._data)

    @property
    def exists(self):
        return self.path.exists

    @property
    def only_read(self) -> bool:
        return self._only_read

    @classmethod
    def add(cls, *args, **kwargs):
        raise NotImplementedError("it's not implemented")

    @classmethod
    def parse_ext(cls, ext: str) -> str:
        if not isinstance(ext, str):
            raise TypeError("Error during parsing ext. Send a string.")
        return "." + ext.replace(".", "")

    def _get_mode(self, mode) -> str:
        """
        if mode is 'r' return 'r+' or 'r+b' else mode is 'w' return 'w+' or 'w+b'
        """
        return f"{mode}+b" if self._is_byte else f"{mode}+"

    def size(self, unit: str = "KB"):
        """
        returns the size that the file occupies on the disk.

        :param unit: choose anyone in ('B', 'KB', 'MB', 'GB', 'TB')

        """
        assert isinstance(
                unit, str
        ), f"expected {str.__name__} not {type(unit).__name__}."

        unit = unit.upper()

        assert (
                unit in self._size_map
        ), f"{repr(unit)} is'nt valid. Choose anyone in {tuple(self._size_map)}"
        # subtracts the size of the python object
        return (
                       memory_of_this(self._data) - self._data.__class__().__sizeof__()
               ) / self._size_map[unit]

    def _load(self, mode=None, **kwargs) -> Any:
        if self._path.suffix in self._dont_read:
            logger.warning(
                    f"I can't read this file. See class attribute <{self.__class__.__name__}._dont_read>"
            )
            return
        mode = mode or self._get_mode("r")
        encoding = None if self._is_byte else "utf-8"
        assert "r" in mode, f"{mode} for read isn't valid."
        try:
            with open(self._path.path, mode=mode, encoding=encoding, **kwargs) as fp:
                parsed = self._parse_fp(fp)
                self._last_update = self.updated_at
                return parsed
        except PermissionError as err:
            logger.error(err)
            return
        except Exception as err:
            raise IOError(f"Error when trying to read the file {self._path}\n{err}")

    def save(
            self,
            on_new_path: Union[str, Path] = None,
            mode=None,
            exist_ok=False,
            overwrite=False,
            force=False,
            encoding=None,
            **kwargs,
    ):

        if not force:
            assert self._only_read is False, f"{self.name} is only read."

            if on_new_path is not None:
                self.set_path(on_new_path)

            if self.was_changed:
                assert overwrite is True, AssertionError(
                        f"File change detected (last change {self.updated_at}), if you want to "
                        + f"overwrite set overwrite=True"
                )

            assert (
                    (exist_ok or not self.path.exists) or overwrite is True or force is True
            ), FileExistsError(
                    "File exists. If you want override, please send 'exist_ok=True'"
            )

        mode = mode or self._get_mode("w")
        assert "w" in mode, f"{mode} for write isn't valid."
        encoding = None if self._is_byte else "utf-8"
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_file = Path(tmpdirname).join(self.name)
            if not self._newline and not self._is_byte:
                kwargs.update({"newline": ""})
            try:
                with open(tmp_file, mode=mode, encoding=encoding, **kwargs) as fp:
                    self._save_fp(fp)
            except Exception as err:
                raise IOError(f"Error writing to the file system file: {err}")
            tmp_file.cp(self.path)
            self._last_update = self.updated_at

    def delete(self):
        self._path.rm()
        self._is_deleted = True

    def set_path(self, path_):
        p = Path(path_)
        assert not p.is_dir, f"{p.path} is a directory."
        self._path = p

    def undo(self):
        if self._current_change > 0:
            index = -2 if self._current_change == len(self._change_history) else -1
            self._select_change(index)

    def redo(self):
        if self._current_change < len(self._change_history):
            self._select_change(+1)

    def sample_items(self, k: int = 1, is_random=False):
        """
        Get sample of this.

        @param k: length of sample
        @param is_random: bool
        """
        assert isinstance(k, int) and k > 0, f"k = {k} isn't valid."
        return sample(self.data, k=k, is_random=is_random)

    def __len__(self):
        if self._length is None:
            return len(self._data)
        return self._length

    def __next__(self):
        for i in self._data:
            yield i

    def __iter__(self):
        return self.__next__()

    @classmethod
    def load(cls, path_, **kwargs):
        return cls(path_, creation_mode=False, **kwargs)

    @classmethod
    def create(cls, path_, data=None, **kwargs):
        return cls(path_=path_, data=data, creation_mode=True, **kwargs)


class _Generic(_FileIO):
    _is_byte: bool = True
    _only_read = False
    _ext_allowed = ()
    _split_lines = False

    def _parse_fp(self, fp: BytesIO) -> List[bytes]:
        if self._split_lines or not bool(self.ext):
            data = fp.read().splitlines()
        else:
            data = fp.read()
        return self.parse(data)

    def _save_fp(self, fp: Union[TextIO, BytesIO]) -> None:
        fp.write(
                self._convert("\n").join(
                        i if isinstance(i, (str, bytes)) else str(i) for i in self._data
                )
        )

    def _convert(self, val):
        if self._is_byte and not isinstance(val, bytes):
            return str(val).encode()
        if not self._is_byte and isinstance(val, str):
            return string_to_literal(val)
        return val

    def parse(self, data) -> List[bytes]:
        if isinstance(data, (bytes, str, int, float, complex)):
            return [self._convert(data)]
        if not data:
            return [self._convert("")]
        if isinstance(data, (list, tuple)):
            return [self._convert(i) for i in data]
        raise ValueError(f"{type(data)} isn't valid.")

    def add(self, data, line=-1):
        data = self.parse(data)
        if line != -1:
            for pos, i in enumerate(data, line):
                self._data.insert(pos, i)
        else:
            self._data += data
            return
        self._set_change("_data", self._data.copy())

    def remove(self, line: Union[int, str]):
        self._data.pop(line)
        self._set_change("_data", self._data.copy())


class _TxtIO(_Generic):
    _is_byte: bool = False
    _only_read = False
    _ext_allowed = ()
    _split_lines = True


class _JsonIO(_FileIO):
    _is_byte: bool = False
    _only_read = False
    _ext_allowed = (".json",)
    indent = False
    ensure_ascii = False

    def __init__(self, path_: Path, indent=False, string_eval=True, **kwargs):
        self.indent = indent
        self.string_eval = string_eval
        if "ensure_ascii" in kwargs:
            self.ensure_ascii = kwargs.pop("ensure_ascii")
        super().__init__(path_, **kwargs)

    @staticmethod
    def _object_hook(obj):
        return {string_to_literal(k): string_to_literal(v) for k, v in obj.items()}

    def _parse_fp(self, fp: TextIO) -> dict:
        return json.load(
                fp, object_hook=self._object_hook if self.string_eval else None
        )

    def _save_fp(self, fp):
        json.dump(
                self._data,
                fp,
                indent=4 if self.indent else None,
                ensure_ascii=self.ensure_ascii,
        )

    def items(self) -> ItemsView:
        return self._data.items()

    def keys(self) -> KeysView:
        return self._data.keys()

    def values(self) -> ValuesView:
        return self._data.values()

    def parse(self, data) -> Union[dict, list]:
        if isinstance(data, (dict, list)):
            return data
        elif isinstance(data, (str, bytes)):
            return json.loads(
                    data, object_hook=self._object_hook if self.string_eval else None
            )
        else:
            raise TypeError(f"{type(data)} isn't valid.")

    def _add(self, k, v):
        self._data[k] = v

    def add(self, key, value):
        self._add(key, value)

    def remove(self, key):
        self._data.pop(key)

    def get(self, _key):
        return self._data.get(_key)

    def __getitem__(self, item):
        try:
            return self.data[item]
        except KeyError:
            raise KeyError(f"{item} not found.")

    def __setitem__(self, key, value):
        self.add(key, value)

    def __iter__(self):
        for i in self.data:
            yield i


class _Mp4IO(_FileIO, ABC):
    _is_byte: bool = True
    _only_read = True
    _ext_allowed = (".mp4",)

    def _save_fp(self, fp):
        fp.write(self._data)

    def _parse_fp(self, fp: BytesIO) -> bytes:
        return fp.read()

    def parse(self, data):
        return self._data


class _CsvIO(_FileIO):
    _is_byte: bool = False
    _only_read = False
    _newline = False
    _ext_allowed = (".csv",)

    def __init__(
            self,
            *args,
            cols: Union[Tuple[str], List[str]] = (),
            fill_with=None,
            str_to_literal=True,
            has_col=False,
            **kwargs,
    ):
        """
        :param data: expected row or list of rows with `n` elements equal to `k` cols, if not it will be fill with
        `fill_with` value.
        :param fill_with: fill empty
        :param has_col: bool
        :param literal_values: Convert `str` on literal types e.g `"'1'"` (str) -> `1` (int)
        """
        self._cols = cols
        self._str_to_literal = str_to_literal
        self._fill_with = fill_with
        self._n_values = 0
        self._has_col = has_col
        super().__init__(*args, **kwargs)

    @property
    def cols(self):
        return tuple(
                value if value is not None else f"unamed_col_{i}"
                for i, value in enumerate(self._cols)
        )

    @property
    def n_cols(self):
        return len(self.cols)

    @property
    def n_values(self):
        return self._n_values

    @property
    def rows(self):
        """
        generator for each row

        format -> {col_name: value, col_name2: value, ...}

        :return: rows with column. Is a dict type
        """
        for row in self.data:
            yield dict(zip(self.cols, row))

    def _parse_fp(self, fp: TextIO, **kwargs) -> List[List[Any]]:
        reader = csv.reader(fp)
        return self.parse(reader)

    def _save_fp(self, fp: TextIO):
        writer = csv.DictWriter(fp, fieldnames=self.cols)
        writer.writeheader()
        writer.writerows(self.rows)

    def _normalize_row(self, row, fill_with=None, literal_values=True):
        if not is_sequence(row):
            row = [row]
        if self._cols:
            row = (
                fill(value=list(row), max_size=len(self.cols), with_=fill_with)
                if len(row) < len(self.cols)
                else row
            )
        return [
            string_to_literal(item) if isinstance(item, str) else item
            for item in row
            if literal_values
        ]

    def parse(
            self,
            data: Union[Iterable[Iterable[Any]], str, int, float, complex],
            fill_with=None,
    ) -> List[List[Any]]:
        if fill_with is None:
            fill_with = self._fill_with
        normalized_data = []
        if isinstance(data, dict):
            self._cols = tuple(data)
            data = data.values()
        if not is_sequence(data):
            data = [data]
        for row in data:
            if not self._cols:
                # set cols on first iter
                self._cols = fill(value=list(row), max_size=len(row), with_=fill_with)
                continue
            row_normalized = self._normalize_row(
                    row, fill_with=fill_with, literal_values=self._str_to_literal
            )
            self._n_values += len(row_normalized)
            normalized_data.append(row_normalized)
        return normalized_data

    def to_dict(self):
        return dict(zip(self.cols, get_cols(self.data)))

    def flatten(self):
        return flatten(self._data)

    def shape(self):
        return get_shape(self._data)

    def _add(self, row: List[Any], index, fill_with=None):
        row = self.parse([row], fill_with=fill_with)
        if index != -1:
            for pos, i in enumerate(row, index):
                self._data.insert(pos, i)
        else:
            self._data += row

    def add(self, row: List[Any], index=-1, fill_with=None):
        """
        Add row is similar to list.append

        :param row: expected row with n elements equal to k cols, if not it will be fill with `fill_with` value.
        :param index: position that will be added
        :param fill_with: None
        """
        self._add(row=row, index=index, fill_with=fill_with)

    def __getitem__(self, item):
        if isinstance(item, str):
            if item not in self.cols:
                raise ValueError(f"Column name <{item}> not in {self.cols}")
            return get_cols(self.data)[self.cols.index(item)]
        if isinstance(item, tuple):
            lines, col = item
            if isinstance(col, tuple):
                raise ValueError("Isn't Possible.")
            assert col <= self.n_cols - 1, ValueError(
                    f"Invalid Column. Choice available index {list(range(self.n_cols))}"
            )
            return get_cols(self.data)[col][lines]
        return super().__getitem__(item)


class _PyObj(_FileIO):
    _is_byte: bool = True
    _only_read = False
    _newline = False
    _ext_allowed = (".pkl",)

    def _parse_fp(self, fp: TextIO, **kwargs) -> List[List[Any]]:
        loaded_obj = pickle.load(fp, **kwargs)
        return self.parse(loaded_obj)

    def _save_fp(self, fp: TextIO):
        pickle.dump(self.data, fp)

    def parse(self, data):
        return data


class _ZipFileIO(_FileIO):
    _is_byte: bool = False
    _only_read = False
    _newline = False
    _ext_allowed = (".zip",)

    def __init__(self, path_: Path, **kwargs):

        super().__init__(path_, **kwargs)

    def add(self, data: Union[str, list, tuple]):
        parsed = self.parse(data)
        self._data += parsed

    def _parse_fp(self, fp: Union[TextIO, BytesIO]) -> Any:
        return self.unzip(fp.name)

    @classmethod
    def unzip(
            cls,
            file_path,
            save_on: str = None,
            members: List = None,
            k=None,
            is_random=False,
    ):
        with ZipFile(file_path, mode="r") as myzip:
            members = members or sample(myzip.namelist(), k, is_random)
            if save_on:
                mkdir(save_on)
            myzip.extractall(
                    save_on, members=Progress.prog(list(members), name="Extracting")
            )

    def _save_fp(self, fp: Union[TextIO, BytesIO]) -> None:
        raise NotImplementedError

    @classmethod
    def load(cls, path_, load_on_memory=False, **kwargs):
        return cls(
                path_=path_, creation_mode=False, load_on_memory=load_on_memory, **kwargs
        )

    def save(
            self,
            on_new_path: Union[str, Path] = None,
            exist_ok=False,
            overwrite=False,
            force=False,
            **kwargs,
    ):
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_file = Path(tmpdirname).join(self.name)
            try:
                with ZipFile(
                        tmp_file.path, mode="w", compression=ZIP_DEFLATED
                ) as myzip:
                    for i in self.data:
                        myzip.write(i)
            except Exception as err:
                raise IOError(f"Error writing to the file system file: {err}")
            tmp_file.cp(self.path)
            self._last_update = self.updated_at

    def parse(self, data):
        if not data:
            return []
        if isinstance(data, (str, Path)):
            data = [data]

        if not is_sequence(data):
            raise ValueError("Send list of paths invalid.")
        res = []

        for p in data:
            p = Path(p)
            if not p.exists:
                raise ValueError(f"{p.path} not Found.")
            res.append(p)
        return res


class _SrtFile(_TxtIO):
    _only_read = False
    _ext_allowed = (".srt", ".vtt")

    @classmethod
    def is_block_start(cls, line) -> bool:
        return bool(
                re.match(r"\d{2}:\d{2}:\d{2}[.,]\d+\s-->\s\d{2}:\d{2}:\d{2}[.,]\d+", line)
        )

    def parse(self, data):
        data = super().parse(data)
        blocks = []
        current_block = 0
        block = []
        block_started = False
        for i in data:
            if self.is_block_start(i):
                if block_started:
                    blocks.append(self.Block(current_block, *block))
                block_started = True
                current_block += 1
                block = []
            if block_started and not i.isdigit():
                block.append(i)
        if block:
            blocks.append(self.Block(current_block, *block))

        return blocks

    @property
    def string(self):
        return "\n".join(
                [str(block_item) for block in self.data for block_item in block]
        )

    def add(self, data, line=-1):
        raise NotImplementedError

    @classmethod
    def from_url(cls, url: str, *args, **kwargs):
        # TODO: need improves
        resp = request.get(url, *args, **kwargs)
        if resp.code == 200:
            return cls(Path("./subtitle.srt"), resp.data, creation_mode=True)

    @property
    def text(self):
        return [
            content.strip()
            for block in self.data
            for content in block.content
            if content
        ]

    class Block:
        def __init__(self, number, *content):
            self.number = number
            time_, *content = content

            self.time = time_
            start, time_separator, end = time_.split()
            self._start = start
            self._end = end
            self.content = content
            self._time_separator = time_separator

        @property
        def start(self):
            return self._start

        @property
        def end(self):
            return self._end

        @property
        def _values(self):
            return (self.number, self.time, *self.content)

        def __getitem__(self, item):
            return self._values[item]

        def __repr__(self):
            return f"{self._values}"


class FileIO:
    txt = _TxtIO
    json = _JsonIO
    mp4 = _Mp4IO
    csv = _CsvIO
    pkl = _PyObj
    zip = _ZipFileIO
    srt = _SrtFile
    _generic = _Generic

    def __new__(cls, path_, data=None, **kwargs):
        raise Exception(
                f"Cannot instantiate {cls.__name__}, use the methods ['create', 'load', 'load_files']"
        )

    @classmethod
    def create(cls, path_: Union[Type[Path], str], data: Any, **kwargs) -> _FileIO:
        path_ = Path(path_)
        return cls.lookup(path_.suffix).create(path_=path_, data=data, **kwargs)

    @classmethod
    def lookup(cls, ext: str) -> Type[_FileIO]:
        ext = ext.replace(".", "")
        if hasattr(cls, ext):
            return getattr(cls, ext)
        return cls._generic

    @classmethod
    def load(cls, path_: Union[str, Path], **kwargs) -> _FileIO:
        path_ = Path(path_)
        if not path_.exists:
            raise FileNotFoundError(f"{path_.path} not found.")
        return cls.lookup(path_.suffix).load(path_=path_, **kwargs)

    @classmethod
    def load_from_clipboard(cls):
        return cls.load(Path(clipboard()))

    @classmethod
    def load_files(
            cls,
            path_,
            ext=None,
            contains_in_name: List = (),
            not_contains_in_name=(),
            take_empty=True,
            recursive=False,
            ignore_dirs=(),
    ):

        ext = ext or ""
        f_paths = Path(path_).list_files(
                ext=ext,
                contains_in_name=contains_in_name,
                not_contains_in_name=not_contains_in_name,
                recursive=recursive,
                ignore_dirs=ignore_dirs,
        )
        loaded = []
        for p in f_paths:
            file_ = cls.load(path_=p)
            if file_ is None:
                continue
            if take_empty is True and file_.is_empty:
                continue
            loaded.append(file_)
        return loaded


FileType = Type[_Generic]
FileTxtType = Type[_TxtIO]
FileJsonType = Type[_JsonIO]
FileCsvType = Type[_CsvIO]
FilePythonObj = Type[_PyObj]



"""
Copyright (c) 2019 The Cereja Project
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import sys
from typing import Union, AnyStr
import unicodedata

__all__ = ["Unicode"]


class Unicode(object):
    """
    Copyright (c) 2019 The Cereja Project
    source: https://github.com/jlsneto/cereja

    Returns the unicode that the `value` represents.
    Supports unicodes in the `range(1114111)`

    usage e.g:

    Acceptable string and integer values
    (binary, octal, decimal or hex)
    >>> Unicode("0x1F352") # Unicode("1F352") is acceptable
    Unicode('\U0001f352', 'CHERRIES')
    >>> Unicode(127826) # Unicode(0x1F352) is acceptable
    Unicode('\U0001f352', 'CHERRIES')

    Acceptable operators '+' and '-'
    >>> unicode_a = Unicode(127826)
    Unicode('\U0001f352', 'CHERRIES')
    >>> unicode_b = unicode_a + 10
    Unicode('\U0001f35c', 'STEAMING BOWL')
    >>> unicode_c = unicode_a - 50
    Unicode('\U0001f320', 'SHOOTING STAR')
    :param value: you can sent strings and integers
    :return: Type Unicode
    """

    __max_unicode = sys.maxunicode

    def __init__(self, value: Union[str, int]):
        self._value = self.__parse(value)
        self.value = self.__value
        self.name = self.__name
        self.decimal = self.__decimal
        self.hex = self.__hex
        self.bin = self.__bin
        self.oct = self.__oct

    def __repr__(self):
        return f"{self.__class__.__name__}{self._value, self.name}"

    def __str__(self):
        return self.value

    def __add__(self, other):
        if isinstance(other, int):
            other = self.parse(other)
            return self.parse(self.decimal + other.decimal)
        raise TypeError("Isn't possible. You can add with integer type")

    def __sub__(self, other):
        if isinstance(other, int):
            other = self.parse(other)
            return self.parse(self.decimal - other.decimal)
        raise TypeError("Isn't possible. You can sub with integer type")

    @classmethod
    def __parse_ord(cls, value: str) -> int:
        value = value.encode("ascii", "backslashreplace").lower()

        if value.startswith(b"\u") or value.startswith(b"\\u"):
            value = value.replace(b"\u", b"")

        value = value.decode("unicode_escape").lower()
        if value.startswith("u"):
            value = value.replace("u+", "").replace("u", "")

        if value.startswith("0x"):
            value = int(value, 16)
        elif value.startswith("0b"):
            value = int(value, 2)
        elif value.startswith("0o"):
            value = int(value, 8)
        else:
            value = int(value, 16)
        return value

    @classmethod
    def __parse(cls, code_) -> AnyStr:
        try:
            if isinstance(code_, str):
                code_ = cls.__parse_ord(code_)
            return chr(code_)
        except Exception as err:
            msg_error = "Conversion error."
            if isinstance(code_, int):
                if code_ > cls.__max_unicode:
                    msg_error = f"{msg_error} Max unicode is {hex(cls.__max_unicode)} ({cls.__max_unicode} on base 10)"
            raise ValueError(msg_error) from err

    @classmethod
    def parse(cls, value: Union[str, int]) -> "Unicode":
        """
        Copyright (c) 2019 The Cereja Project
        source: https://github.com/jlsneto/cereja

        Returns the unicode that the `value` represents.
        Supports unicodes in the `range(1114111)`

        usage e.g:

        Acceptable string values
        >>> Unicode.parse("1F352") # is the unicode value
        '\U0001f352'
        >>> Unicode.parse("0x1F352") # full hexadecimal
        '\U0001f352'

        Acceptable int values in system of numerical notation:
        (binary, octal, decimal or hex)
        >>> Unicode.parse(127826) # is a decimal
        '\U0001f352'
        >>> Unicode.parse(0x1F352) # is a same value in hexadecimal
        '\U0001f352'

        :param value: you can sent strings and integers
        :return: Type Unicode
        """
        return cls(value)

    @property
    def __decimal(self):
        return ord(self._value)

    @property
    def __hex(self):
        return hex(self.__decimal)

    @property
    def __oct(self):
        return oct(self.__decimal)

    @property
    def __bin(self):
        return bin(self.__decimal)

    @property
    def __name(self):
        try:
            return unicodedata.name(self._value)
        except ValueError:
            return "UNKNOWN"

    @property
    def __value(self) -> str:
        return self._value



import tempfile
import unittest

from cereja.mltools import preprocess

from cereja.system import Path
from cereja.mltools.preprocess import separate
from cereja.mltools.data import Tokenizer
from cereja.mltools.data import Freq


class BaseTest(unittest.TestCase):
    def loop_on_examples_pairs(self, func, examples):
        msg_ = f"Error on {func.__name__}"
        for example, expected_response, args, kwargs in examples:
            if args and kwargs:
                self.assertEqual(
                    func(example, *args, **kwargs), expected_response, msg=msg_
                )
            elif args:
                self.assertEqual(func(example, *args), expected_response, msg=msg_)
            elif kwargs:
                self.assertEqual(func(example, **kwargs), expected_response, msg=msg_)
            else:
                self.assertEqual(func(example), expected_response, msg=msg_)


class TokenizerTest(unittest.TestCase):
    data = None

    def test_sanity(self):
        with tempfile.TemporaryDirectory() as tempdir:
            to_encode = "oi tudo tranquilo"
            tokenizer = Tokenizer(
                ["oi amigo tudo bem"], preprocess_function=separate, use_unk=False
            )

            self.assertEqual(tokenizer.last_index, 13)
            encoded = tokenizer.encode([to_encode])
            valid_encode_tokens = [
                [
                    tokenizer._item_to_index["oi"],
                    tokenizer._item_to_index["tudo"],
                    tokenizer._item_to_index["tranquilo"],
                ]
            ]
            self.assertEqual(tokenizer.last_index, 14)
            self.assertEqual(encoded, valid_encode_tokens)
            decoded = tokenizer.decode(encoded[0])
            self.assertEqual(decoded, to_encode)
            temp_file = Path(tempdir).join("tokenizer.json")
            tokenizer.to_json(temp_file)

            # Valid.
            tokenizer = Tokenizer.load_from_json(temp_file)
            encoded = tokenizer.encode([to_encode])
            self.assertEqual(tokenizer.last_index, 14)
            self.assertEqual(encoded, valid_encode_tokens)
            decoded = tokenizer.decode(encoded[0])
            self.assertEqual(decoded, to_encode)


class PreprocessModuleTest(BaseTest):
    def test_sanity(self):
        self.assertEqual(preprocess.separate("how are you?", sep="?"), "how are you ?")
        self.assertEqual(
            preprocess.separate("how are you,man?", sep=("?", ","), between_char=True),
            "how are you , man ?",
        )
        self.assertEqual(preprocess.separate("how are! you?"), "how are ! you ?")

    def test_separate(self):
        examples = (
            ("how are you?", "how are you ?", (), {"sep": "?"}),
            (
                "how are you,man?",
                "how are you , man ?",
                (),
                {"sep": ("?", ","), "between_char": True},
            ),
            ("how are! you?", "how are ! you ?", (), {}),
        )
        self.loop_on_examples_pairs(preprocess.separate, examples)

    def test_accent_remove(self):
        examples = (
            (
                "olá tudo bem amigo? Você é um cara legal!",
                "ola tudo bem amigo? Voce e um cara legal!",
                (),
                {},
            ),
            ("isso é muito difícil", "isso e muito dificil", (), {}),
        )
        self.loop_on_examples_pairs(preprocess.accent_remove, examples)

    def test_replace_english_contractions(self):
        examples = (("i'll", "i shall", (), {}),)
        self.loop_on_examples_pairs(preprocess.replace_english_contractions, examples)

    def test_remove_extra_chars(self):
        examples = (("i   like it!!!", "i like it!", (), {}),)
        self.loop_on_examples_pairs(preprocess.remove_extra_chars, examples)

    def test_remove_non_language_elements(self):
        examples = (("\U0001F352 i like it!", "i like it!", (), {}),)
        self.loop_on_examples_pairs(preprocess.remove_non_language_elements, examples)

    def test_remove_punctuation(self):
        examples = (
            ("\U0001F352 i like it!", "\U0001F352 i like it", (), {}),
            (
                "olá tudo bem amigo?! Você é um cara legal!",
                "olá tudo bem amigo Você é um cara legal",
                (),
                {},
            ),
            ("e aí? quero R$ 40,00", "e aí quero R 4000", (), {}),
        )
        self.loop_on_examples_pairs(preprocess.remove_punctuation, examples)

        examples2 = (
            ("\U0001F352 i like it!", "\U0001F352 i like it", (), {}),
            (
                "olá tudo bem amigo?! Você é um cara legal!",
                "olá tudo bem amigo Você é um cara legal",
                (),
                {},
            ),
            ("e aí? quero R$ 40,00", "e aí quero R$ 40,00", (), {"punctuation": "?!"}),
        )
        self.loop_on_examples_pairs(preprocess.remove_punctuation, examples2)

    def test_remove_stop_words(self):
        examples = (
            ("my name is Joab", "my name Joab", (), {}),
            ("you are cool!", "you cool!", (), {}),
        )
        self.loop_on_examples_pairs(preprocess.remove_stop_words, examples)


class FreqTest(unittest.TestCase):
    def setUp(self) -> None:
        self.freq = Freq([1, 2, 3, 4, 4])

    def test_probability(self):
        self.assertEqual(self.freq.probability, {1: 0.2, 2: 0.2, 3: 0.2, 4: 0.4})

    def test_item_prob(self):
        self.assertEqual(self.freq.item_prob(4), 0.4)

    def test_sanity(self):
        self.assertEqual(preprocess.separate("how are you?", sep="?"), "how are you ?")
        self.assertEqual(
            preprocess.separate("how are you,man?", sep=("?", ","), between_char=True),
            "how are you , man ?",
        )
        self.assertEqual(preprocess.separate("how are! you?"), "how are ! you ?")

        freq = Freq([1, 1, 2, 2, 3, 3, 4, 5, 6, 7, 8, "hi", "o", "a"])
        self.assertEqual(
            freq.sample(max_freq=1),
            {4: 1, 5: 1, 6: 1, 7: 1, 8: 1, "hi": 1, "o": 1, "a": 1},
        )
        self.assertEqual(freq.sample(freq=2), {1: 2, 3: 2, 2: 2})

        self.assertRaises(AssertionError, freq.sample, freq=1, max_freq=2)
        self.assertRaises(AssertionError, freq.sample, freq=1, min_freq=2)

        freq = Freq([1, 2, 3, 3, 4, 5, 6, 7, 6, 7, 12, 31, 123, 5, 3])
        self.assertEqual(
            freq.least_freq(),
            {123: 1, 31: 1, 12: 1, 4: 1, 2: 1, 1: 1, 7: 2, 6: 2, 5: 2, 3: 3},
        )


if __name__ == "__main__":
    unittest.main()
